from functools import total_ordering
from itertools import chain
from os.path import splitext, basename
from typing import Dict, Optional, List, Iterable, Set, Union, TypeVar, Generic, Callable, Sequence, Type

from geopy import Point

from betty.locale import Localized, Datey
from betty.media_type import MediaType
from betty.path import extension

T = TypeVar('T')


class EventHandlingSetList(Generic[T]):
    def __init__(self, addition_handler: Callable[[T], None], removal_handler: Callable[[T], None]):
        self._values = []
        self._addition_handler = addition_handler
        self._removal_handler = removal_handler

    @property
    def list(self) -> List[T]:
        return list(self._values)

    def prepend(self, *values: T) -> None:
        for value in reversed(values):
            if value in self._values:
                continue
            self._values.insert(0, value)
            self._addition_handler(value)

    def append(self, *values: T) -> None:
        for value in values:
            if value in self._values:
                continue
            self._values.append(value)
            self._addition_handler(value)

    def remove(self, *values: T) -> None:
        for value in values:
            if value not in self._values:
                return
            self._values.remove(value)
            self._removal_handler(value)

    def replace(self, *values: T) -> None:
        self.remove(*list(self._values))
        self.append(*values)

    def clear(self) -> None:
        self.replace()

    def __iter__(self):
        return self._values.__iter__()

    def __len__(self):
        return len(self._values)

    def __getitem__(self, item):
        return self._values[item]


ManyAssociation = Union[EventHandlingSetList[T], Sequence[T]]


class _to_many:
    def __init__(self, self_name: str, associated_name: str):
        self._self_name = self_name
        self._associated_name = associated_name

    def __call__(self, cls):
        _decorated_self_name = '_%s' % self._self_name
        original_init = cls.__init__

        def _init(decorated_self, *args, **kwargs):
            association = EventHandlingSetList(self._create_addition_handler(decorated_self),
                                               self._create_removal_handler(decorated_self))
            setattr(decorated_self, _decorated_self_name, association)
            original_init(decorated_self, *args, **kwargs)
        cls.__init__ = _init
        setattr(cls, self._self_name, property(
            lambda decorated_self: getattr(decorated_self, _decorated_self_name),
            lambda decorated_self, values: getattr(decorated_self, _decorated_self_name).replace(*values),
            lambda decorated_self: getattr(decorated_self, _decorated_self_name).clear(),
        ))
        return cls

    def _create_addition_handler(self, decorated_self):
        raise NotImplementedError

    def _create_removal_handler(self, decorated_self):
        raise NotImplementedError


class many_to_many(_to_many):
    def _create_addition_handler(self, decorated_self):
        return lambda associated: getattr(associated, self._associated_name).append(decorated_self)

    def _create_removal_handler(self, decorated_self):
        return lambda associated: getattr(associated, self._associated_name).remove(decorated_self)


def bridged_many_to_many(left_associated_name: str, left_self_name: str, right_self_name: str, right_associated_name: str):
    def decorator(cls):
        cls = many_to_one(left_self_name, left_associated_name, lambda decorated_self: delattr(decorated_self, right_self_name))(cls)
        cls = many_to_one(right_self_name, right_associated_name, lambda decorated_self: delattr(decorated_self, left_self_name))(cls)
        return cls
    return decorator


class one_to_many(_to_many):
    def _create_addition_handler(self, decorated_self):
        return lambda associated: setattr(associated, self._associated_name, decorated_self)

    def _create_removal_handler(self, decorated_self):
        return lambda associated: setattr(associated, self._associated_name, None)


def many_to_one(self_name: str, associated_name: str, _removal_handler: Optional[Callable[[T], None]] = None):
    def decorator(cls):
        _decorated_self_name = '_%s' % self_name
        original_init = cls.__init__

        def _init(decorated_self, *args, **kwargs):
            association = None
            setattr(decorated_self, _decorated_self_name, association)
            original_init(decorated_self, *args, **kwargs)
        cls.__init__ = _init

        def _set(decorated_self, value):
            previous_value = getattr(decorated_self, _decorated_self_name)
            if previous_value == value:
                return
            setattr(decorated_self, _decorated_self_name, value)
            if previous_value is not None:
                getattr(previous_value, associated_name).remove(decorated_self)
                if value is None and _removal_handler is not None:
                    _removal_handler(decorated_self)
            if value is not None:
                getattr(value, associated_name).append(decorated_self)
        setattr(cls, self_name, property(
            lambda decorated_self: getattr(decorated_self, _decorated_self_name),
            _set,
            lambda decorated_self: _set(decorated_self, None),
        ))
        return cls
    return decorator


class Resource:
    @classmethod
    def resource_type_name(cls) -> str:
        raise NotImplementedError


class HasPrivacy:
    private: Optional[bool]

    def __init__(self):
        self.private = None


class Dated:
    date: Optional[Datey]

    def __init__(self):
        self.date = None


class Note:
    text: str

    def __init__(self, text: str):
        self._text = text

    @property
    def text(self):
        return self._text


class HasNotes:
    notes: List[Note]

    def __init__(self):
        self.notes = []


class Identifiable:
    def __init__(self, identifiable_id: str):
        self._id = identifiable_id

    @property
    def id(self) -> str:
        return self._id


class Described:
    description: Optional[str]

    def __init__(self):
        self.description = None


class HasMediaType:
    media_type: Optional[MediaType]

    def __init__(self):
        self.media_type = None


class Link(HasMediaType, Localized, Described):
    url: str
    relationship: Optional[str]
    label: Optional[str]

    def __init__(self, url: str):
        HasMediaType.__init__(self)
        Localized.__init__(self)
        Described.__init__(self)
        self.url = url
        self.label = None
        self.relationship = None


class HasLinks:
    def __init__(self):
        self._links = set()

    @property
    def links(self) -> Set[Link]:
        return self._links


@many_to_many('resources', 'files')
class File(Resource, Identifiable, Described, HasPrivacy, HasMediaType, HasNotes):
    resources: ManyAssociation['HasFiles']
    notes: List[Note]

    def __init__(self, file_id: str, path: str, media_type: Optional[MediaType] = None):
        Identifiable.__init__(self, file_id)
        Described.__init__(self)
        HasPrivacy.__init__(self)
        HasMediaType.__init__(self)
        HasNotes.__init__(self)
        self._path = path
        self.media_type = media_type

    @classmethod
    def resource_type_name(cls) -> str:
        return 'file'

    @property
    def path(self) -> str:
        return self._path

    @property
    def name(self) -> str:
        return basename(self._path)

    @property
    def basename(self) -> str:
        return splitext(self._path)[0]

    @property
    def extension(self) -> Optional[str]:
        return extension(self._path)

    @property
    def sources(self) -> Iterable['Source']:
        for resource in self.resources:
            if isinstance(resource, Source):
                yield resource
            if isinstance(resource, Citation):
                yield resource.source

    @property
    def citations(self) -> Iterable['Citation']:
        for resource in self.resources:
            if isinstance(resource, Citation):
                yield resource


@many_to_many('files', 'resources')
class HasFiles:
    files: ManyAssociation[File]

    @property
    def associated_files(self) -> Sequence[File]:
        return self.files


@many_to_one('contained_by', 'contains')
@one_to_many('contains', 'contained_by')
@one_to_many('citations', 'source')
class Source(Resource, Dated, HasFiles, HasLinks, HasPrivacy):
    name: Optional[str]
    contained_by: 'Source'
    contains: ManyAssociation['Source']
    citations: ManyAssociation['Citation']
    author: Optional[str]
    publisher: Optional[str]

    def __init__(self, name: Optional[str] = None):
        Dated.__init__(self)
        HasFiles.__init__(self)
        HasLinks.__init__(self)
        HasPrivacy.__init__(self)
        self.name = name
        self.author = None
        self.publisher = None

    @classmethod
    def resource_type_name(cls) -> str:
        return 'source'


class IdentifiableSource(Source, Identifiable):
    def __init__(self, source_id: str, *args, **kwargs):
        Identifiable.__init__(self, source_id)
        Source.__init__(self, *args, **kwargs)


@many_to_many('facts', 'citations')
@many_to_one('source', 'citations')
class Citation(Resource, Dated, HasFiles, HasPrivacy):
    facts: ManyAssociation['HasCitations']
    source: Source
    location: Optional[str]

    def __init__(self, source: Source):
        Dated.__init__(self)
        HasFiles.__init__(self)
        HasPrivacy.__init__(self)
        self.location = None
        self.source = source

    @classmethod
    def resource_type_name(cls) -> str:
        return 'citation'


class IdentifiableCitation(Citation, Identifiable):
    def __init__(self, citation_id: str, *args, **kwargs):
        Identifiable.__init__(self, citation_id)
        Citation.__init__(self, *args, **kwargs)


@many_to_many('citations', 'facts')
class HasCitations:
    citations: ManyAssociation[Citation]


class PlaceName(Localized, Dated):
    def __init__(self, name: str, locale: Optional[str] = None, date: Optional[Datey] = None):
        Localized.__init__(self)
        self._name = name
        self.locale = locale
        self.date = date

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self._name == other._name and self.locale == other.locale

    def __repr__(self):
        return '<%s.%s(%s, %s)>' % (self.__class__.__module__, self.__class__.__name__, self.name, repr(self.locale))

    def __str__(self):
        return self._name

    @property
    def name(self) -> str:
        return self._name


@bridged_many_to_many('enclosed_by', 'encloses', 'enclosed_by', 'encloses')
class Enclosure(Dated, HasCitations):
    encloses: 'Place'
    enclosed_by: 'Place'

    def __init__(self, encloses: 'Place', enclosed_by: 'Place'):
        Dated.__init__(self)
        HasCitations.__init__(self)
        self.encloses = encloses
        self.enclosed_by = enclosed_by


@one_to_many('events', 'place')
@one_to_many('enclosed_by', 'encloses')
@one_to_many('encloses', 'enclosed_by')
class Place(Resource, Identifiable, HasLinks):
    enclosed_by: ManyAssociation[Enclosure]
    encloses: ManyAssociation[Enclosure]

    def __init__(self, place_id: str, names: List[PlaceName]):
        Identifiable.__init__(self, place_id)
        HasLinks.__init__(self)
        self._names = names
        self._coordinates = None

    @classmethod
    def resource_type_name(cls) -> str:
        return 'place'

    @property
    def names(self) -> List[PlaceName]:
        return self._names

    @property
    def coordinates(self) -> Optional[Point]:
        return self._coordinates

    @coordinates.setter
    def coordinates(self, coordinates: Point):
        self._coordinates = coordinates


class PresenceRole:
    @classmethod
    def name(cls) -> str:
        raise NotImplementedError

    @property
    def label(self) -> str:
        raise NotImplementedError


class Subject(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'subject'

    @property
    def label(self) -> str:
        return _('Subject')


class Witness(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'witness'

    @property
    def label(self) -> str:
        return _('Witness')


class Beneficiary(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'beneficiary'

    @property
    def label(self) -> str:
        return _('Beneficiary')


class Attendee(PresenceRole):
    @classmethod
    def name(cls) -> str:
        return 'attendee'

    @property
    def label(self) -> str:
        return _('Attendee')


@bridged_many_to_many('presences', 'person', 'event', 'presences')
class Presence:
    person: Optional['Person']
    event: Optional['Event']
    role: PresenceRole

    def __init__(self, person: 'Person', role: PresenceRole, event: 'Event'):
        self.person = person
        self.role = role
        self.event = event


class EventType:
    @classmethod
    def name(cls) -> str:
        raise NotImplementedError

    @property
    def label(self) -> str:
        raise NotImplementedError

    @classmethod
    def comes_before(cls) -> Set[Type['EventType']]:
        return set()

    @classmethod
    def comes_after(cls) -> Set[Type['EventType']]:
        return set()


class UnknownEventType(EventType):
    @classmethod
    def name(cls) -> str:
        return 'unknown'

    @property
    def label(self) -> str:
        return _('Unknown')


class DerivableEventType(EventType):
    pass  # pragma: no cover


class CreatableDerivableEventType(DerivableEventType):
    pass  # pragma: no cover


class PreBirthEventType(EventType):
    @classmethod
    def comes_before(cls) -> Set[Type['EventType']]:
        return {Birth}


class LifeEventType(EventType):
    @classmethod
    def comes_after(cls) -> Set[Type['EventType']]:
        return {Birth}

    @classmethod
    def comes_before(cls) -> Set[Type['EventType']]:
        return {Death}


class PostDeathEventType(EventType):
    @classmethod
    def comes_after(cls) -> Set[Type['EventType']]:
        return {Death}


class Birth(CreatableDerivableEventType):
    @classmethod
    def name(cls) -> str:
        return 'birth'

    @property
    def label(self) -> str:
        return _('Birth')

    @classmethod
    def comes_before(cls) -> Set[Type[EventType]]:
        return {LifeEventType}


class Baptism(LifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'baptism'

    @property
    def label(self) -> str:
        return _('Baptism')


class Adoption(LifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'adoption'

    @property
    def label(self) -> str:
        return _('Adoption')


class Death(CreatableDerivableEventType):
    @classmethod
    def name(cls) -> str:
        return 'death'

    @property
    def label(self) -> str:
        return _('Death')

    @classmethod
    def comes_after(cls) -> Set[Type[EventType]]:
        return {LifeEventType}


class Funeral(PostDeathEventType, DerivableEventType):
    @classmethod
    def name(cls) -> str:
        return 'funeral'

    @property
    def label(self) -> str:
        return _('Funeral')

    @classmethod
    def comes_after(cls) -> Set[Type[EventType]]:
        return {Death}


class FinalDispositionEventType(PostDeathEventType, DerivableEventType):
    @classmethod
    def comes_after(cls) -> Set[Type[EventType]]:
        return {Death}


class Cremation(FinalDispositionEventType):
    @classmethod
    def name(cls) -> str:
        return 'cremation'

    @property
    def label(self) -> str:
        return _('Cremation')


class Burial(FinalDispositionEventType):
    @classmethod
    def name(cls) -> str:
        return 'burial'

    @property
    def label(self) -> str:
        return _('Burial')


class Will(PostDeathEventType):
    @classmethod
    def name(cls) -> str:
        return 'will'

    @property
    def label(self) -> str:
        return _('Will')

    @classmethod
    def comes_after(cls) -> Set[Type[EventType]]:
        return {Death}


class Engagement(LifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'engagement'

    @property
    def label(self) -> str:
        return _('Engagement')

    @classmethod
    def comes_before(cls) -> Set[Type[EventType]]:
        return {Marriage}


class Marriage(LifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'marriage'

    @property
    def label(self) -> str:
        return _('Marriage')


class MarriageAnnouncement(LifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'marriage-announcement'

    @property
    def label(self) -> str:
        return _('Announcement of marriage')

    @classmethod
    def comes_before(cls) -> Set[Type[EventType]]:
        return {Marriage}


class Divorce(LifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'divorce'

    @property
    def label(self) -> str:
        return _('Divorce')

    @classmethod
    def comes_after(cls) -> Set[Type[EventType]]:
        return {Marriage}


class DivorceAnnouncement(LifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'divorce-announcement'

    @property
    def label(self) -> str:
        return _('Announcement of divorce')

    @classmethod
    def comes_after(cls) -> Set[Type[EventType]]:
        return {Marriage}

    @classmethod
    def comes_before(cls) -> Set[Type[EventType]]:
        return {Divorce}


class Residence(LifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'residence'

    @property
    def label(self) -> str:
        return _('Residence')


class Immigration(LifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'immigration'

    @property
    def label(self) -> str:
        return _('Immigration')


class Emigration(LifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'emigration'

    @property
    def label(self) -> str:
        return _('Emigration')


class Occupation(LifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'occupation'

    @property
    def label(self) -> str:
        return _('Occupation')


class Retirement(LifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'retirement'

    @property
    def label(self) -> str:
        return _('Retirement')


class Correspondence(EventType):
    @classmethod
    def name(cls) -> str:
        return 'correspondence'

    @property
    def label(self) -> str:
        return _('Correspondence')


class Confirmation(LifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'confirmation'

    @property
    def label(self) -> str:
        return _('Confirmation')


class Missing(LifeEventType):
    @classmethod
    def name(cls) -> str:
        return 'missing'

    @property
    def label(self) -> str:
        return _('Missing')


EVENT_TYPE_TYPES = [
    Birth,
    Baptism,
    Adoption,
    Death,
    Funeral,
    Cremation,
    Burial,
    Will,
    Engagement,
    Marriage,
    MarriageAnnouncement,
    Divorce,
    DivorceAnnouncement,
    Residence,
    Immigration,
    Emigration,
    Occupation,
    Retirement,
    Correspondence,
    Confirmation,
]


@many_to_one('place', 'events')
@one_to_many('presences', 'event')
class Event(Resource, Dated, HasFiles, HasCitations, Described, HasPrivacy):
    place: Place
    presences: ManyAssociation[Presence]

    def __init__(self, event_type: EventType, date: Optional[Datey] = None):
        Dated.__init__(self)
        HasFiles.__init__(self)
        HasCitations.__init__(self)
        Described.__init__(self)
        HasPrivacy.__init__(self)
        self.date = date
        self._type = event_type

    @classmethod
    def resource_type_name(cls) -> str:
        return 'event'

    def __repr__(self):
        return '<%s.%s(%s, date=%s)>' % (self.__class__.__module__, self.__class__.__name__, repr(self.type), repr(self.date))

    @property
    def type(self):
        return self._type

    @property
    def associated_files(self) -> Sequence[File]:
        files = [
            *self.files,
            *[file for citation in self.citations for file in citation.associated_files],
        ]
        # Preserve the original order.
        seen = set()
        for file in files:
            if file in seen:
                continue
            seen.add(file)
            yield file


class IdentifiableEvent(Event, Identifiable):
    def __init__(self, event_id: str, *args, **kwargs):
        Identifiable.__init__(self, event_id)
        Event.__init__(self, *args, **kwargs)


@total_ordering
@many_to_one('person', 'names')
class PersonName(Localized, HasCitations):
    person: Optional['Person']

    def __init__(self, individual: Optional[str] = None, affiliation: Optional[str] = None):
        Localized.__init__(self)
        HasCitations.__init__(self)
        self._individual = individual
        self._affiliation = affiliation

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, PersonName):
            return NotImplemented
        return (self._affiliation or '', self._individual or '') == (other._affiliation or '', other._individual or '')

    def __gt__(self, other):
        if other is None:
            return True
        if not isinstance(other, PersonName):
            return NotImplemented
        return (self._affiliation or '', self._individual or '') > (other._affiliation or '', other._individual or '')

    @property
    def individual(self) -> str:
        return self._individual

    @property
    def affiliation(self) -> str:
        return self._affiliation


@total_ordering
@many_to_many('parents', 'children')
@many_to_many('children', 'parents')
@one_to_many('presences', 'person')
@one_to_many('names', 'person')
class Person(Resource, Identifiable, HasFiles, HasCitations, HasLinks, HasPrivacy):
    parents: ManyAssociation['Person']
    children: ManyAssociation['Person']
    presences: ManyAssociation[Presence]
    names: ManyAssociation[PersonName]

    def __init__(self, person_id: str):
        Identifiable.__init__(self, person_id)
        HasFiles.__init__(self)
        HasCitations.__init__(self)
        HasLinks.__init__(self)
        HasPrivacy.__init__(self)

    @classmethod
    def resource_type_name(cls) -> str:
        return 'person'

    def __eq__(self, other):
        if not isinstance(other, Person):
            return NotImplemented
        return self.id == other.id

    def __gt__(self, other):
        if not isinstance(other, Person):
            return NotImplemented
        return self.id > other.id

    @property
    def name(self) -> Optional[PersonName]:
        try:
            return self._names.list[0]
        except IndexError:
            return None

    @property
    def alternative_names(self) -> List[PersonName]:
        return self.names.list[1:]

    @property
    def start(self) -> Optional[Event]:
        for event_type in [Birth, Baptism]:
            for presence in self.presences:
                if isinstance(presence.event.type, event_type) and isinstance(presence.role, Subject):
                    return presence.event
        return None

    @property
    def end(self) -> Optional[Event]:
        for event_type in [Death, Burial]:
            for presence in self.presences:
                if isinstance(presence.event.type, event_type) and isinstance(presence.role, Subject):
                    return presence.event
        return None

    @property
    def siblings(self) -> List:
        siblings = []
        for parent in self._parents:
            for sibling in parent.children:
                if sibling != self and sibling not in siblings:
                    siblings.append(sibling)
        return siblings

    @property
    def associated_files(self) -> Sequence[File]:
        files = [
            *self.files,
            *[file for name in self.names for citation in name.citations for file in citation.associated_files],
            *[file for presence in self.presences for file in presence.event.associated_files]
        ]
        # Preserve the original order.
        seen = set()
        for file in files:
            if file in seen:
                continue
            seen.add(file)
            yield file


RESOURCE_TYPES = [
    Citation,
    Event,
    File,
    Person,
    Place,
    Source,
]


class Ancestry:
    files: Dict[str, File]
    people: Dict[str, Person]
    places: Dict[str, Place]
    events: Dict[str, IdentifiableEvent]
    sources: Dict[str, IdentifiableSource]
    citations: Dict[str, IdentifiableCitation]

    def __init__(self):
        self.files = {}
        self.people = {}
        self.places = {}
        self.events = {}
        self.sources = {}
        self.citations = {}

    @property
    def resources(self) -> Iterable[Resource]:
        return chain(
            self.files.values(),
            self.people.values(),
            self.places.values(),
            self.events.values(),
            self.sources.values(),
            self.citations.values(),
        )



import cgi
from typing import Dict, Optional, List

EXTENSIONS = {
    'text/html': 'html',
    'application/json': 'json',
}


class InvalidMediaType(ValueError):
    pass


class MediaType:
    def __init__(self, media_type: str):
        self._str = media_type
        type_part, self._parameters = cgi.parse_header(media_type)
        try:
            self._type, self._subtype = type_part.split('/')
            if not self._subtype:
                raise ValueError('The subtype must not be empty.')
        except ValueError:
            raise InvalidMediaType('"%s" is not a valid media type.', media_type)

    @property
    def type(self) -> str:
        return self._type

    @property
    def subtype(self) -> str:
        return self._subtype

    @property
    def subtypes(self) -> List[str]:
        return self._subtype.split('+')[0].split('.')

    @property
    def suffix(self) -> Optional[str]:
        if '+' not in self._subtype:
            return None

        return self._subtype.split('+')[-1]

    @property
    def parameters(self) -> Dict[str, str]:
        return self._parameters

    def __str__(self):
        return self._str

    def __eq__(self, other):
        if not isinstance(other, MediaType):
            return NotImplemented
        return (self.type, self.subtype, self.parameters) == (other.type, other.subtype, other.parameters)



import os
from os import mkdir, path
from tempfile import TemporaryDirectory

from betty.fs import iterfiles, FileSystem, hashfile
from betty.asyncio import sync
from betty.tests import TestCase


class IterfilesTest(TestCase):
    @sync
    async def test_iterfiles(self):
        with TemporaryDirectory() as working_directory_path:
            working_subdirectory_path = path.join(working_directory_path, 'subdir')
            mkdir(working_subdirectory_path)
            open(path.join(working_directory_path, 'rootfile'), 'a').close()
            open(path.join(working_directory_path, '.hiddenrootfile'), 'a').close()
            open(path.join(working_subdirectory_path, 'subdirfile'), 'a').close()
            actual = [actualpath[len(working_directory_path) + 1:] async for actualpath in iterfiles(working_directory_path)]
        expected = [
            '.hiddenrootfile',
            'rootfile',
            path.join('subdir', 'subdirfile'),
        ]
        self.assertCountEqual(expected, actual)


class HashfileTest(TestCase):
    def test_hashfile_with_identical_file(self):
        file_path = path.join(path.dirname(path.dirname(__file__)), 'assets', 'public', 'static', 'betty-16x16.png')
        self.assertEquals(hashfile(file_path), hashfile(file_path))

    def test_hashfile_with_different_files(self):
        file_path_1 = path.join(path.dirname(path.dirname(__file__)), 'assets', 'public', 'static', 'betty-16x16.png')
        file_path_2 = path.join(path.dirname(path.dirname(__file__)), 'assets', 'public', 'static', 'betty-512x512.png')
        self.assertNotEquals(hashfile(file_path_1), hashfile(file_path_2))


class FileSystemTest(TestCase):
    @sync
    async def test_open(self):
        with TemporaryDirectory() as source_path_1:
            with TemporaryDirectory() as source_path_2:
                with open(path.join(source_path_1, 'apples'), 'w') as f:
                    f.write('apples')
                with open(path.join(source_path_2, 'apples'), 'w') as f:
                    f.write('notapples')
                with open(path.join(source_path_1, 'oranges'), 'w') as f:
                    f.write('oranges')
                with open(path.join(source_path_2, 'bananas'), 'w') as f:
                    f.write('bananas')

                sut = FileSystem(source_path_1, source_path_2)

                with await sut.open('apples') as f:
                    self.assertEquals('apples', f.read())
                with await sut.open('oranges') as f:
                    self.assertEquals('oranges', f.read())
                with await sut.open('bananas') as f:
                    self.assertEquals('bananas', f.read())

                with self.assertRaises(FileNotFoundError):
                    with await sut.open('mangos'):
                        pass

    @sync
    async def test_open_with_first_file_path_alternative_first_source_path(self):
        with TemporaryDirectory() as source_path_1:
            with TemporaryDirectory() as source_path_2:
                with open(path.join(source_path_1, 'pinkladies'), 'w') as f:
                    f.write('pinkladies')
                with open(path.join(source_path_2, 'pinkladies'), 'w') as f:
                    f.write('notpinkladies')
                with open(path.join(source_path_1, 'apples'), 'w') as f:
                    f.write('notpinkladies')
                with open(path.join(source_path_2, 'apples'), 'w') as f:
                    f.write('notpinkladies')

                sut = FileSystem(source_path_1, source_path_2)

                with await sut.open('pinkladies', 'apples') as f:
                    self.assertEquals('pinkladies', f.read())

    @sync
    async def test_open_with_first_file_path_alternative_second_source_path(self):
        with TemporaryDirectory() as source_path_1:
            with TemporaryDirectory() as source_path_2:
                with open(path.join(source_path_2, 'pinkladies'), 'w') as f:
                    f.write('pinkladies')
                with open(path.join(source_path_1, 'apples'), 'w') as f:
                    f.write('notpinkladies')
                with open(path.join(source_path_2, 'apples'), 'w') as f:
                    f.write('notpinkladies')

                sut = FileSystem(source_path_1, source_path_2)

                with await sut.open('pinkladies', 'apples') as f:
                    self.assertEquals('pinkladies', f.read())

    @sync
    async def test_open_with_second_file_path_alternative_first_source_path(self):
        with TemporaryDirectory() as source_path_1:
            with TemporaryDirectory() as source_path_2:
                with open(path.join(source_path_1, 'apples'), 'w') as f:
                    f.write('apples')
                with open(path.join(source_path_2, 'apples'), 'w') as f:
                    f.write('notapples')

                sut = FileSystem(source_path_1, source_path_2)

                with await sut.open('pinkladies', 'apples') as f:
                    self.assertEquals('apples', f.read())

    @sync
    async def test_copy2(self):
        with TemporaryDirectory() as source_path_1:
            with TemporaryDirectory() as source_path_2:
                with open(path.join(source_path_1, 'apples'), 'w') as f:
                    f.write('apples')
                with open(path.join(source_path_2, 'apples'), 'w') as f:
                    f.write('notapples')
                with open(path.join(source_path_1, 'oranges'), 'w') as f:
                    f.write('oranges')
                with open(path.join(source_path_2, 'bananas'), 'w') as f:
                    f.write('bananas')

                with TemporaryDirectory() as destination_path:
                    sut = FileSystem(source_path_1, source_path_2)

                    await sut.copy2('apples', destination_path)
                    await sut.copy2('oranges', destination_path)
                    await sut.copy2('bananas', destination_path)

                    with await sut.open(path.join(destination_path, 'apples')) as f:
                        self.assertEquals('apples', f.read())
                    with await sut.open(path.join(destination_path, 'oranges')) as f:
                        self.assertEquals('oranges', f.read())
                    with await sut.open(path.join(destination_path, 'bananas')) as f:
                        self.assertEquals('bananas', f.read())

                    with self.assertRaises(FileNotFoundError):
                        await sut.copy2('mangos', destination_path)

    @sync
    async def test_copytree(self):
        with TemporaryDirectory() as source_path_1:
            os.makedirs(path.join(source_path_1, 'basket'))
            with TemporaryDirectory() as source_path_2:
                os.makedirs(path.join(source_path_2, 'basket'))
                with open(path.join(source_path_1, 'basket', 'apples'), 'w') as f:
                    f.write('apples')
                with open(path.join(source_path_2, 'basket', 'apples'), 'w') as f:
                    f.write('notapples')
                with open(path.join(source_path_1, 'basket', 'oranges'), 'w') as f:
                    f.write('oranges')
                with open(path.join(source_path_2, 'basket', 'bananas'), 'w') as f:
                    f.write('bananas')

                with TemporaryDirectory() as destination_path:
                    sut = FileSystem(source_path_1, source_path_2)

                    await sut.copytree('', destination_path)

                    with await sut.open(path.join(destination_path, 'basket', 'apples')) as f:
                        self.assertEquals('apples', f.read())
                    with await sut.open(path.join(destination_path, 'basket', 'oranges')) as f:
                        self.assertEquals('oranges', f.read())
                    with await sut.open(path.join(destination_path, 'basket', 'bananas')) as f:
                        self.assertEquals('bananas', f.read())



from textwrap import indent
from typing import List


class ContextError(Exception):

    """
    An error with a stack of contextual messages.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._contexts: List[str] = []

    def __str__(self):
        return super().__str__() + '\n' + indent('\n'.join(self._contexts), '- ')

    def add_context(self, context: str):
        """
        Add a message describing the error's context.

        :param context: str
        :return: ExternalContextError
        """
        self._contexts.append(context)
        return self


class UserFacingError(Exception):
    """
    A user-facing error.

    This type of error is fatal, but fixing it does not require knowledge of Betty's internals or the stack trace
    leading to the error. It must therefore have an end-user-friendly message, and its stack trace must not be shown.
    """
    pass



from datetime import datetime
from tempfile import TemporaryDirectory
from typing import Optional

from parameterized import parameterized

from betty.ancestry import Person, Presence, Event, Source, IdentifiableSource, File, \
    IdentifiableCitation, Subject, Attendee, Birth, Marriage, Death
from betty.config import Configuration
from betty.asyncio import sync
from betty.locale import Date, DateRange
from betty.parse import parse
from betty.plugin.privatizer import Privatizer, privatize_event, privatize_source, privatize_citation, privatize_person
from betty.site import Site
from betty.tests import TestCase


def _expand_person(generation: int):
    lifetime_threshold = 125
    multiplier = abs(generation) + 1 if generation < 0 else 1
    lifetime_threshold_year = datetime.now().year - lifetime_threshold * multiplier
    date_under_lifetime_threshold = Date(lifetime_threshold_year + 1, 1, 1)
    date_range_start_under_lifetime_threshold = DateRange(date_under_lifetime_threshold)
    date_range_end_under_lifetime_threshold = DateRange(None, date_under_lifetime_threshold)
    date_over_lifetime_threshold = Date(lifetime_threshold_year - 1, 1, 1)
    date_range_start_over_lifetime_threshold = DateRange(date_over_lifetime_threshold)
    date_range_end_over_lifetime_threshold = DateRange(None, date_over_lifetime_threshold)
    return parameterized.expand([
        # If there are no events for a person, they are private.
        (True, None, None),
        (True, True, None),
        (False, False, None),

        # Deaths and other end-of-life events are special, but only for the person whose privacy is being checked:
        # - If they're present without dates, the person isn't private.
        # - If they're present and their dates or date ranges' end dates are in the past, the person isn't private.
        (generation != 0, None, Event(Death(), date=Date(datetime.now().year, datetime.now().month, datetime.now().day))),
        (generation != 0, None, Event(Death(), date=date_under_lifetime_threshold)),
        (True, None, Event(Death(), date=date_range_start_under_lifetime_threshold)),
        (generation != 0, None, Event(Death(), date=date_range_end_under_lifetime_threshold)),
        (False, None, Event(Death(), date=date_over_lifetime_threshold)),
        (True, None, Event(Death(), date=date_range_start_over_lifetime_threshold)),
        (False, None, Event(Death(), date=date_range_end_over_lifetime_threshold)),
        (True, True, Event(Death())),
        (False, False, Event(Death())),
        (generation != 0, None, Event(Death())),

        # Regular events without dates do not affect privacy.
        (True, None, Event(Birth())),
        (True, True, Event(Birth())),
        (False, False, Event(Birth())),

        # Regular events with incomplete dates do not affect privacy.
        (True, None, Event(Birth(), date=Date())),
        (True, True, Event(Birth(), date=Date())),
        (False, False, Event(Birth(), date=Date())),

        # Regular events under the lifetime threshold do not affect privacy.
        (True, None, Event(Birth(), date=date_under_lifetime_threshold)),
        (True, True, Event(Birth(), date=date_under_lifetime_threshold)),
        (False, False, Event(Birth(), date=date_under_lifetime_threshold)),
        (True, None, Event(Birth(), date=date_range_start_under_lifetime_threshold)),
        (True, True, Event(Birth(), date=date_range_start_under_lifetime_threshold)),
        (False, False, Event(Birth(), date=date_range_start_under_lifetime_threshold)),
        (True, None, Event(Birth(), date=date_range_end_under_lifetime_threshold)),
        (True, True, Event(Birth(), date=date_range_end_under_lifetime_threshold)),
        (False, False, Event(Birth(), date=date_range_end_under_lifetime_threshold)),

        # Regular events over the lifetime threshold affect privacy.
        (False, None, Event(Birth(), date=date_over_lifetime_threshold)),
        (True, True, Event(Birth(), date=date_over_lifetime_threshold)),
        (False, False, Event(Birth(), date=date_over_lifetime_threshold)),
        (True, None, Event(Birth(), date=date_range_start_over_lifetime_threshold)),
        (True, True, Event(Birth(), date=date_range_start_over_lifetime_threshold)),
        (False, False, Event(Birth(), date=date_range_start_over_lifetime_threshold)),
        (False, None, Event(Birth(), date=date_range_end_over_lifetime_threshold)),
        (True, True, Event(Birth(), date=date_range_end_over_lifetime_threshold)),
        (False, False, Event(Birth(), date=date_range_end_over_lifetime_threshold)),
    ])


class PrivatizerTest(TestCase):
    @sync
    async def test_post_parse(self):
        person = Person('P0')
        Presence(person, Subject(), Event(Birth()))

        source_file = File('F0', __file__)
        source = IdentifiableSource('S0', 'The Source')
        source.private = True
        source.files.append(source_file)

        citation_file = File('F0', __file__)
        citation_source = Source('The Source')
        citation = IdentifiableCitation('C0', citation_source)
        citation.private = True
        citation.files.append(citation_file)

        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.plugins[Privatizer] = None
            async with Site(configuration) as site:
                site.ancestry.people[person.id] = person
                site.ancestry.sources[source.id] = source
                site.ancestry.citations[citation.id] = citation
                await parse(site)

            self.assertTrue(person.private)
            self.assertTrue(source_file.private)
            self.assertTrue(citation_file.private)

    def test_privatize_person_should_not_privatize_if_public(self):
        source_file = File('F0', __file__)
        source = Source('The Source')
        source.files.append(source_file)
        citation_file = File('F1', __file__)
        citation = IdentifiableCitation('C0', source)
        citation.files.append(citation_file)
        event_as_subject = Event(Birth())
        event_as_attendee = Event(Marriage())
        person_file = File('F2', __file__)
        person = Person('P0')
        person.private = False
        person.citations.append(citation)
        person.files.append(person_file)
        Presence(person, Subject(), event_as_subject)
        Presence(person, Attendee(), event_as_attendee)
        privatize_person(person, 125)
        self.assertEqual(False, person.private)
        self.assertIsNone(citation.private)
        self.assertIsNone(source.private)
        self.assertIsNone(person_file.private)
        self.assertIsNone(citation_file.private)
        self.assertIsNone(source_file.private)
        self.assertIsNone(event_as_subject.private)
        self.assertIsNone(event_as_attendee.private)

    def test_privatize_person_should_privatize_if_private(self):
        source_file = File('F0', __file__)
        source = Source('The Source')
        source.files.append(source_file)
        citation_file = File('F1', __file__)
        citation = IdentifiableCitation('C0', source)
        citation.files.append(citation_file)
        event_as_subject = Event(Birth())
        event_as_attendee = Event(Marriage())
        person_file = File('F2', __file__)
        person = Person('P0')
        person.private = True
        person.citations.append(citation)
        person.files.append(person_file)
        Presence(person, Subject(), event_as_subject)
        Presence(person, Attendee(), event_as_attendee)
        privatize_person(person, 125)
        self.assertTrue(person.private)
        self.assertTrue(citation.private)
        self.assertTrue(source.private)
        self.assertTrue(person_file.private)
        self.assertTrue(citation_file.private)
        self.assertTrue(source_file.private)
        self.assertTrue(event_as_subject.private)
        self.assertIsNone(event_as_attendee.private)

    @_expand_person(0)
    def test_privatize_person_without_relatives(self, expected, private, event: Optional[Event]):
        person = Person('P0')
        person.private = private
        if event is not None:
            Presence(person, Subject(), event)
        privatize_person(person, 125)
        self.assertEquals(expected, person.private)

    @_expand_person(1)
    def test_privatize_person_with_child(self, expected, private, event: Optional[Event]):
        person = Person('P0')
        person.private = private
        child = Person('P1')
        if event is not None:
            Presence(child, Subject(), event)
        person.children.append(child)
        privatize_person(person, 125)
        self.assertEquals(expected, person.private)

    @_expand_person(2)
    def test_privatize_person_with_grandchild(self, expected, private, event: Optional[Event]):
        person = Person('P0')
        person.private = private
        child = Person('P1')
        person.children.append(child)
        grandchild = Person('P2')
        if event is not None:
            Presence(grandchild, Subject(), event)
        child.children.append(grandchild)
        privatize_person(person, 125)
        self.assertEquals(expected, person.private)

    @_expand_person(3)
    def test_privatize_person_with_great_grandchild(self, expected, private, event: Optional[Event]):
        person = Person('P0')
        person.private = private
        child = Person('P1')
        person.children.append(child)
        grandchild = Person('P2')
        child.children.append(grandchild)
        great_grandchild = Person('P2')
        if event is not None:
            Presence(great_grandchild, Subject(), event)
        grandchild.children.append(great_grandchild)
        privatize_person(person, 125)
        self.assertEquals(expected, person.private)

    @_expand_person(-1)
    def test_privatize_person_with_parent(self, expected, private, event: Optional[Event]):
        person = Person('P0')
        person.private = private
        parent = Person('P1')
        if event is not None:
            Presence(parent, Subject(), event)
        person.parents.append(parent)
        privatize_person(person, 125)
        self.assertEquals(expected, person.private)

    @_expand_person(-2)
    def test_privatize_person_with_grandparent(self, expected, private, event: Optional[Event]):
        person = Person('P0')
        person.private = private
        parent = Person('P1')
        person.parents.append(parent)
        grandparent = Person('P2')
        if event is not None:
            Presence(grandparent, Subject(), event)
        parent.parents.append(grandparent)
        privatize_person(person, 125)
        self.assertEquals(expected, person.private)

    @_expand_person(-3)
    def test_privatize_person_with_great_grandparent(self, expected, private, event: Optional[Event]):
        person = Person('P0')
        person.private = private
        parent = Person('P1')
        person.parents.append(parent)
        grandparent = Person('P2')
        parent.parents.append(grandparent)
        great_grandparent = Person('P2')
        if event is not None:
            Presence(great_grandparent, Subject(), event)
        grandparent.parents.append(great_grandparent)
        privatize_person(person, 125)
        self.assertEquals(expected, person.private)

    def test_privatize_event_should_not_privatize_if_public(self):
        source_file = File('F0', __file__)
        source = Source('The Source')
        source.files.append(source_file)
        citation_file = File('F1', __file__)
        citation = IdentifiableCitation('C0', source)
        citation.files.append(citation_file)
        event_file = File('F1', __file__)
        event = Event(Birth())
        event.private = False
        event.citations.append(citation)
        event.files.append(event_file)
        person = Person('P0')
        Presence(person, Subject(), event)
        privatize_event(event)
        self.assertEqual(False, event.private)
        self.assertIsNone(event_file.private)
        self.assertIsNone(citation.private)
        self.assertIsNone(source.private)
        self.assertIsNone(citation_file.private)
        self.assertIsNone(source_file.private)
        self.assertIsNone(person.private)

    def test_privatize_event_should_privatize_if_private(self):
        source_file = File('F0', __file__)
        source = Source('The Source')
        source.files.append(source_file)
        citation_file = File('F1', __file__)
        citation = IdentifiableCitation('C0', source)
        citation.files.append(citation_file)
        event_file = File('F1', __file__)
        event = Event(Birth())
        event.private = True
        event.citations.append(citation)
        event.files.append(event_file)
        person = Person('P0')
        Presence(person, Subject(), event)
        privatize_event(event)
        self.assertTrue(event.private)
        self.assertTrue(event_file.private)
        self.assertTrue(citation.private)
        self.assertTrue(source.private)
        self.assertTrue(citation_file.private)
        self.assertTrue(source_file.private)
        self.assertIsNone(person.private)

    def test_privatize_source_should_not_privatize_if_public(self):
        file = File('F0', __file__)
        source = IdentifiableSource('S0', 'The Source')
        source.private = False
        source.files.append(file)
        privatize_source(source)
        self.assertEqual(False, source.private)
        self.assertIsNone(file.private)

    def test_privatize_source_should_privatize_if_private(self):
        file = File('F0', __file__)
        source = IdentifiableSource('S0', 'The Source')
        source.private = True
        source.files.append(file)
        privatize_source(source)
        self.assertTrue(source.private)
        self.assertTrue(file.private)

    def test_privatize_citation_should_not_privatize_if_public(self):
        source_file = File('F0', __file__)
        source = Source('The Source')
        source.files.append(source_file)
        citation_file = File('F1', __file__)
        citation = IdentifiableCitation('C0', source)
        citation.private = False
        citation.files.append(citation_file)
        privatize_citation(citation)
        self.assertEqual(False, citation.private)
        self.assertIsNone(source.private)
        self.assertIsNone(citation_file.private)
        self.assertIsNone(source_file.private)

    def test_privatize_citation_should_privatize_if_private(self):
        source_file = File('F0', __file__)
        source = Source('The Source')
        source.files.append(source_file)
        citation_file = File('F1', __file__)
        citation = IdentifiableCitation('C0', source)
        citation.private = True
        citation.files.append(citation_file)
        privatize_citation(citation)
        self.assertTrue(citation.private)
        self.assertTrue(source.private)
        self.assertTrue(citation_file.private)
        self.assertTrue(source_file.private)



from subprocess import CompletedProcess, CalledProcessError

from betty import subprocess
from betty.tests import TestCase


class RunTest(TestCase):
    def test_without_errors(self):
        process = subprocess.run(['true'])
        self.assertIsInstance(process, CompletedProcess)

    def test_with_errors(self):
        with self.assertRaises(CalledProcessError):
            subprocess.run(['false'])



import hashlib
import os
import shutil
from collections import deque
from contextlib import suppress
from os import walk, path
from os.path import join, dirname, exists, relpath, getmtime
from shutil import copy2
from tempfile import mkdtemp
from typing import AsyncIterable


async def iterfiles(path: str) -> AsyncIterable[str]:
    for dir_path, _, filenames in walk(path):
        for filename in filenames:
            yield join(dir_path, filename)


def makedirs(path: str) -> None:
    os.makedirs(path, 0o755, True)


def hashfile(path: str) -> str:
    return hashlib.md5(':'.join([str(getmtime(path)), path]).encode('utf-8')).hexdigest()


class FileSystem:
    def __init__(self, *paths):
        self._paths = deque(paths)

    @property
    def paths(self) -> deque:
        return self._paths

    async def open(self, *file_paths: str):
        for file_path in file_paths:
            for fs_path in self._paths:
                with suppress(FileNotFoundError):
                    return open(join(fs_path, file_path))
        raise FileNotFoundError

    async def copy2(self, source_path: str, destination_path: str) -> str:
        for fs_path in self._paths:
            with suppress(FileNotFoundError):
                return copy2(join(fs_path, source_path), destination_path)
        tried_paths = [join(fs_path, source_path) for fs_path in self._paths]
        raise FileNotFoundError('Could not find any of %s.' %
                                ', '.join(tried_paths))

    async def copytree(self, source_path: str, destination_path: str) -> str:
        for fs_path in self._paths:
            async for file_source_path in iterfiles(join(fs_path, source_path)):
                file_destination_path = join(destination_path, relpath(
                    file_source_path, join(fs_path, source_path)))
                if not exists(file_destination_path):
                    makedirs(dirname(file_destination_path))
                    copy2(file_source_path, file_destination_path)
        return destination_path


class DirectoryBackup:
    def __init__(self, root_path: str, backup_path: str):
        self._root_path = root_path
        self._backup_path = backup_path

    async def __aenter__(self):
        self._tmp = mkdtemp()
        with suppress(FileNotFoundError):
            shutil.move(path.join(self._root_path, self._backup_path), self._tmp)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        with suppress(FileNotFoundError):
            shutil.move(path.join(self._tmp, self._backup_path),
                        path.join(self._root_path, self._backup_path))
        shutil.rmtree(self._tmp)



from os import path

from voluptuous import Invalid

from betty.importlib import import_any


def Path():
    def _path(v):
        try:
            return path.abspath(path.expanduser(v))
        except TypeError as e:
            raise Invalid(e)

    return _path


def Importable():
    def _importable(v):
        try:
            return import_any(v)
        except ImportError as e:
            raise Invalid(e)

    return _importable



from typing import Callable, Type, List, Any


class TargetedDispatcher:
    def __init__(self, handler_methods):
        self._handler_methods = handler_methods

    async def __call__(self, *args, **kwargs) -> List[Any]:
        return [await handler_method(*args, **kwargs) for handler_method in self._handler_methods]


class Dispatcher:
    def __init__(self):
        self._handlers = []

    def append_handler(self, handler: Callable):
        self._handlers.append(handler)

    def dispatch(self, target_type: Type, target_method_name: str) -> TargetedDispatcher:
        handler_methods = [getattr(handler, target_method_name) for handler in self._handlers if isinstance(handler, target_type)]
        return TargetedDispatcher(handler_methods)



from typing import Dict

from betty.site import Site


class _Resource:
    def __init__(self, name: str, collection_endpoint_summary: str, collection_response_description: str,
                 single_endpoint_summary: str, single_response_description):
        self.name = name
        self.collection_endpoint_summary = collection_endpoint_summary
        self.collection_response_description = collection_response_description
        self.single_endpoint_summary = single_endpoint_summary
        self.single_response_description = single_response_description


def _get_resources():
    return [
        _Resource('file', _('Retrieve the collection of files.'), _(
            'The collection of files.'), _('Retrieve a single file.'), _('The file.')),
        _Resource('person', _('Retrieve the collection of people.'), _(
            'The collection of people.'), _('Retrieve a single person.'), _('The person.')),
        _Resource('place', _('Retrieve the collection of places.'), _(
            'The collection of places.'), _('Retrieve a single place.'), _('The place.')),
        _Resource('event', _('Retrieve the collection of events.'), _(
            'The collection of events.'), _('Retrieve a single event.'), _('The event.')),
        _Resource('citation', _('Retrieve the collection of citations.'), _(
            'The collection of citations.'), _('Retrieve a single citation.'), _('The citation.')),
        _Resource('source', _('Retrieve the collection of sources.'), _(
            'The collection of sources.'), _('Retrieve a single source.'), _('The source.')),
    ]


def build_specification(site: Site) -> Dict:
    specification = {
        'openapi': '3.0.0',
        'servers': [
            {
                'url': site.static_url_generator.generate('/', absolute=True),
            }
        ],
        'info': {
            'title': 'Betty',
            'version': 'dev'
        },
        'paths': {},
        'components': {
            'responses': {
                '401': {
                    'description': _('Unauthorized'),
                    'content': {
                        'application/json': {
                            'schema': {
                                '$ref': '#/components/schemas/betty/error',
                            },
                        },
                    },
                },
                '403': {
                    'description': _('Forbidden'),
                    'content': {
                        'application/json': {
                            'schema': {
                                '$ref': '#/components/schemas/betty/error',
                            },
                        },
                    },
                },
                '404': {
                    'description': _('Not found'),
                    'content': {
                        'application/json': {
                            'schema': {
                                '$ref': '#/components/schemas/betty/error',
                            },
                        },
                    },
                },
            },
            'parameters': {
                'id': {
                    'name': 'id',
                    'in': 'path',
                    'required': True,
                    'description': _('The ID for the resource to retrieve.'),
                    'schema': {
                        'type': 'string',
                    },
                },
            },
            'headers': {
                'Content-Language': {
                    'description': _('An HTTP [Content-Language](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Language) header.'),
                    'schema': {
                        'type': 'string',
                    },
                    'example': site.configuration.default_locale,
                },
            },
            'schemas': {
                'betty': {
                    '$ref': site.static_url_generator.generate('/schema.json#/definitions'),
                },
            },
        },
    }

    # Add resource operations.
    for resource in _get_resources():
        if site.configuration.content_negotiation:
            collection_path = '/%s/' % resource.name
            single_path = '/%s/{id}/' % resource.name
        else:
            collection_path = '/{locale}/%s/index.json' % resource.name
            single_path = '/{locale}/%s/{id}/index.json' % resource.name
        specification['paths'].update({
            collection_path: {
                'get': {
                    'summary': resource.collection_endpoint_summary,
                    'responses': {
                        '200': {
                            'description': resource.collection_response_description,
                            'content': {
                                'application/json': {
                                    'schema': {
                                        '$ref': '#/components/schemas/betty/%sCollection' % resource.name,
                                    },
                                },
                            },
                        },
                    },
                },
            },
            single_path: {
                'get': {
                    'summary': resource.single_endpoint_summary,
                    'responses': {
                        '200': {
                            'description': resource.single_response_description,
                            'content': {
                                'application/json': {
                                    'schema': {
                                        '$ref': '#/components/schemas/betty/%s' % resource.name,
                                    },
                                },
                            },
                        },
                    },
                },
            },
        })

    # Add components for content negotiation.
    if site.configuration.content_negotiation:
        specification['components']['parameters']['Accept-Language'] = {
            'name': 'Accept-Language',
            'in': 'header',
            'description': _('An HTTP [Accept-Language](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept-Language) header.'),
            'schema': {
                'type': 'string',
            },
            'example': site.configuration.locales[site.configuration.default_locale].alias,
        }
        specification['components']['schemas']['html'] = {
            'type': 'string',
            'description': _('An HTML5 document.'),
        }
    else:
        specification['components']['parameters']['locale'] = {
            'name': 'locale',
            'in': 'path',
            'required': True,
            'description': _('A locale name.'),
            'schema': {
                'type': 'string',
                'enum': list(site.configuration.locales.keys())
            },
            'example': site.configuration.locales[site.configuration.default_locale].alias,
        }

    # Add default behavior to all requests.
    for path in specification['paths']:
        specification['paths'][path]['get']['responses'].update({
            '401': {
                '$ref': '#/components/responses/401',
            },
            '403': {
                '$ref': '#/components/responses/403',
            },
            '404': {
                '$ref': '#/components/responses/404',
            },
        })
        if site.configuration.content_negotiation:
            specification['paths'][path]['parameters'] = [
                {
                    '$ref': '#/components/parameters/Accept-Language',
                },
            ]
        else:
            specification['paths'][path]['parameters'] = [
                {
                    '$ref': '#/components/parameters/locale',
                },
            ]

    # Add default behavior to all responses.
    if site.configuration.content_negotiation:
        responses = list(specification['components']['responses'].values())
        for path in specification['paths']:
            responses.append(
                specification['paths'][path]['get']['responses']['200'])
        for response in responses:
            response['content']['text/html'] = {
                'schema': {
                    '$ref': '#/components/schemas/html'
                }
            }
            if 'headers' not in response:
                response['headers'] = {}
            response['headers']['Content-Language'] = {
                '$ref': '#/components/headers/Content-Language',
            }

    return specification



from betty.asyncio import sync
from betty.lock import Locks, AcquiredError
from betty.tests import TestCase


class LocksTest(TestCase):
    @sync
    async def test_acquire(self):
        resource = 999
        sut = Locks()
        sut.acquire(resource)
        with self.assertRaises(AcquiredError):
            sut.acquire(resource)

    @sync
    async def test_release(self):
        resource = 999
        sut = Locks()
        sut.acquire(resource)
        sut.release(resource)
        sut.acquire(resource)



from os.path import join
from tempfile import TemporaryDirectory

from betty.config import Configuration
from betty.asyncio import sync
from betty.generate import generate
from betty.plugin.trees import Trees
from betty.site import Site
from betty.tests import patch_cache, TestCase


class TreesTest(TestCase):
    @patch_cache
    @sync
    async def test_post_render_event(self):
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://ancestry.example.com')
            configuration.mode = 'development'
            configuration.plugins[Trees] = None
            async with Site(configuration) as site:
                await generate(site)
            with open(join(configuration.www_directory_path, 'trees.js')) as f:
                betty_js = f.read()
            self.assertIn('trees.js', betty_js)
            self.assertIn('trees.css', betty_js)
            with open(join(configuration.www_directory_path, 'trees.css')) as f:
                betty_css = f.read()
            self.assertIn('.tree', betty_css)



import logging
from datetime import datetime
from typing import Any, Optional

from betty.ancestry import Ancestry, Person, Event, Citation, Source, HasPrivacy, Subject
from betty.functools import walk
from betty.locale import DateRange, Date
from betty.parse import PostParser
from betty.plugin import Plugin, NO_CONFIGURATION
from betty.site import Site


class Privatizer(Plugin, PostParser):
    def __init__(self, ancestry: Ancestry, lifetime_threshold: int):
        self._ancestry = ancestry
        self._lifetime_threshold = lifetime_threshold

    @classmethod
    def for_site(cls, site: Site, configuration: Any = NO_CONFIGURATION):
        return cls(site.ancestry, site.configuration.lifetime_threshold)

    async def post_parse(self) -> None:
        self.privatize(self._ancestry)

    def privatize(self, ancestry: Ancestry) -> None:
        privatized = 0
        for person in ancestry.people.values():
            private = person.private
            privatize_person(person, self._lifetime_threshold)
            if private is None and person.private is True:
                privatized += 1
        logger = logging.getLogger()
        logger.info('Privatized %d people because they are likely still alive.' % privatized)

        for citation in ancestry.citations.values():
            privatize_citation(citation)

        for source in ancestry.sources.values():
            privatize_source(source)


def _mark_private(has_privacy: HasPrivacy) -> None:
    # Do not change existing explicit privacy declarations.
    if has_privacy.private is None:
        has_privacy.private = True


def privatize_person(person: Person, lifetime_threshold: int) -> None:
    # Do not change existing explicit privacy declarations.
    if person.private is None:
        person.private = _person_is_private(person, lifetime_threshold)

    if not person.private:
        return

    for presence in person.presences:
        if isinstance(presence.role, Subject):
            _mark_private(presence.event)
            privatize_event(presence.event)
    for file in person.files:
        _mark_private(file)
    for citation in person.citations:
        _mark_private(citation)
        privatize_citation(citation)


def privatize_event(event: Event) -> None:
    if not event.private:
        return

    for file in event.files:
        _mark_private(file)
    for citation in event.citations:
        _mark_private(citation)
        privatize_citation(citation)


def privatize_citation(citation: Citation) -> None:
    if not citation.private:
        return

    _mark_private(citation.source)
    privatize_source(citation.source)
    for file in citation.files:
        _mark_private(file)


def privatize_source(source: Source) -> None:
    if not source.private:
        return

    for file in source.files:
        _mark_private(file)


def _person_is_private(person: Person, lifetime_threshold: int) -> bool:
    # A dead person is not private, regardless of when they died.
    if person.end is not None:
        if person.end.date is None:
            return False
        if _event_has_expired(person.end, lifetime_threshold, 0):
            return False

    if _person_has_expired(person, lifetime_threshold, 1):
        return False

    def ancestors(person: Person, generation: int = -1):
        for parent in person.parents:
            yield generation, parent
            yield from ancestors(parent, generation - 1)

    for generation, ancestor in ancestors(person):
        if _person_has_expired(ancestor, lifetime_threshold, abs(generation) + 1):
            return False

    # If any descendant has any expired event, the person is considered not private.
    for descendant in walk(person, 'children'):
        if _person_has_expired(descendant, lifetime_threshold, 1):
            return False

    return True


def _person_has_expired(person: Person, lifetime_threshold: int, multiplier: int) -> bool:
    for presence in person.presences:
        if _event_has_expired(presence.event, lifetime_threshold, multiplier):
            return True
    return False


def _event_has_expired(event: Event, lifetime_threshold: int, multiplier: int) -> bool:
    assert multiplier >= 0

    date = event.date

    if isinstance(date, DateRange):
        # We can only determine event expiration with certainty if we have an end date to work with. Someone born in
        # 2000 can have a valid birth event with a start date of 1800, which does nothing to help us determine
        # expiration.
        date = date.end

    return _date_has_expired(date, lifetime_threshold, multiplier)


def _date_has_expired(date: Optional[Date], lifetime_threshold: int, multiplier: int) -> bool:
    assert multiplier >= 0

    if date is None:
        return False

    if not date.comparable:
        return False

    return date <= Date(datetime.now().year - lifetime_threshold * multiplier, datetime.now().month, datetime.now().day)



from betty.site import Site


class Parser:
    async def parse(self) -> None:
        raise NotImplementedError


class PostParser:
    async def post_parse(self) -> None:
        raise NotImplementedError


async def parse(site: Site) -> None:
    await site.dispatcher.dispatch(Parser, 'parse')()
    await site.dispatcher.dispatch(PostParser, 'post_parse')()



import asyncio
import datetime
import json as stdjson
import os
import re
import warnings
from contextlib import suppress
from os.path import join, relpath
from pathlib import Path
from typing import Union, Dict, Type, Optional, Callable, Iterable, AsyncIterable, Any, Iterator

import pdf2image
from PIL import Image
from PIL.Image import DecompressionBombWarning
from babel import Locale
from geopy import units
from geopy.format import DEGREES_FORMAT
from jinja2 import Environment, select_autoescape, evalcontextfilter, escape, FileSystemLoader, contextfilter, Template
from jinja2.asyncsupport import auto_await
from jinja2.filters import prepare_map, make_attrgetter
from jinja2.nodes import EvalContext
from jinja2.runtime import Macro, resolve_or_missing, StrictUndefined, Context
from jinja2.utils import htmlsafe_json_dumps
from markupsafe import Markup
from resizeimage import resizeimage

from betty.ancestry import File, Citation, Identifiable, Resource, HasLinks, HasFiles, Subject, Witness, Dated, \
    RESOURCE_TYPES
from betty.config import Configuration
from betty.fs import makedirs, hashfile, iterfiles
from betty.functools import walk
from betty.html import HtmlProvider
from betty.importlib import import_any
from betty.json import JSONEncoder
from betty.locale import negotiate_localizeds, Localized, format_datey, Datey, negotiate_locale, Date, DateRange
from betty.lock import AcquiredError
from betty.os import link_or_copy
from betty.path import rootname, extension
from betty.plugin import Plugin
from betty.render import Renderer
from betty.search import Index
from betty.site import Site


class _Plugins:
    def __init__(self, plugins: Dict[Type[Plugin], Plugin]):
        self._plugins = plugins

    def __getitem__(self, plugin_type_name):
        try:
            return self._plugins[import_any(plugin_type_name)]
        except ImportError:
            raise KeyError('Unknown plugin "%s".' % plugin_type_name)

    def __contains__(self, plugin_type_name) -> bool:
        with suppress(ImportError):
            return import_any(plugin_type_name) in self._plugins
        return False


class _Citer:
    def __init__(self):
        self._citations = []

    def __iter__(self) -> Iterable[Citation]:
        return enumerate(self._citations, 1)

    def __len__(self) -> int:
        return len(self._citations)

    def cite(self, citation: Citation) -> int:
        if citation not in self._citations:
            self._citations.append(citation)
        return self._citations.index(citation) + 1

    def track(self) -> None:
        self.clear()

    def clear(self) -> None:
        self._citations = []


class Jinja2Provider:
    @property
    def globals(self) -> Dict[str, Callable]:
        return {}

    @property
    def filters(self) -> Dict[str, Callable]:
        return {}


class BettyEnvironment(Environment):
    site: Site

    def __init__(self, site: Site):
        template_directory_paths = [join(path, 'templates') for path in site.assets.paths]

        Environment.__init__(self,
                             enable_async=True,
                             loader=FileSystemLoader(template_directory_paths),
                             undefined=StrictUndefined,
                             autoescape=select_autoescape(['html']),
                             trim_blocks=True,
                             extensions=[
                                 'jinja2.ext.do',
                                 'jinja2.ext.i18n',
                             ],
                             )

        self.site = site

        if site.configuration.mode == 'development':
            self.add_extension('jinja2.ext.debug')

        self._init_i18n()
        self._init_globals()
        self._init_filters()
        self._init_tests()
        self._init_plugins()

    def _init_i18n(self) -> None:
        # Wrap the callables so they always call the built-ins available runtime, because those change when the current
        # locale does.
        self.install_gettext_callables(
            lambda *args, **kwargs: gettext(*args, **kwargs),
            lambda *args, **kwargs: ngettext(*args, **kwargs),
        )
        self.policies['ext.i18n.trimmed'] = True

    def _init_globals(self) -> None:
        self.globals['site'] = self.site
        self.globals['locale'] = self.site.locale
        today = datetime.date.today()
        self.globals['today'] = Date(today.year, today.month, today.day)
        self.globals['plugins'] = _Plugins(self.site.plugins)
        self.globals['citer'] = _Citer()
        self.globals['search_index'] = lambda: Index(self.site).build()
        self.globals['html_providers'] = list([plugin for plugin in self.site.plugins.values() if isinstance(plugin, HtmlProvider)])
        self.globals['path'] = os.path

    def _init_filters(self) -> None:
        self.filters['set'] = set
        self.filters['unique'] = _filter_unique
        self.filters['map'] = _filter_map
        self.filters['flatten'] = _filter_flatten
        self.filters['walk'] = _filter_walk
        self.filters['locale_get_data'] = lambda locale: Locale.parse(
            locale, '-')
        self.filters['negotiate_localizeds'] = _filter_negotiate_localizeds
        self.filters['sort_localizeds'] = _filter_sort_localizeds
        self.filters['select_localizeds'] = _filter_select_localizeds
        self.filters['negotiate_dateds'] = _filter_negotiate_dateds
        self.filters['select_dateds'] = _filter_select_dateds
        self.filters['json'] = _filter_json
        self.filters['tojson'] = _filter_tojson
        self.filters['paragraphs'] = _filter_paragraphs
        self.filters['format_date'] = _filter_format_date
        self.filters['format_degrees'] = _filter_format_degrees
        self.filters['url'] = _filter_url
        self.filters['static_url'] = self.site.static_url_generator.generate
        self.filters['file'] = lambda *args: _filter_file(self.site, *args)
        self.filters['image'] = lambda *args, **kwargs: _filter_image(self.site, *args, **kwargs)

    def _init_tests(self) -> None:
        self.tests['resource'] = lambda x: isinstance(x, Resource)

        def _build_test_resource_type(resource_type: Type[Resource]):
            def _test_resource(x):
                return isinstance(x, resource_type)
            return _test_resource
        for resource_type in RESOURCE_TYPES:
            self.tests['%s_resource' % resource_type.resource_type_name()] = _build_test_resource_type(resource_type)
        self.tests['identifiable'] = lambda x: isinstance(x, Identifiable)
        self.tests['has_links'] = lambda x: isinstance(x, HasLinks)
        self.tests['has_files'] = lambda x: isinstance(x, HasFiles)
        self.tests['starts_with'] = str.startswith
        self.tests['subject_role'] = lambda x: isinstance(x, Subject)
        self.tests['witness_role'] = lambda x: isinstance(x, Witness)
        self.tests['date_range'] = lambda x: isinstance(x, DateRange)

    def _init_plugins(self) -> None:
        for plugin in self.site.plugins.values():
            if isinstance(plugin, Jinja2Provider):
                self.globals.update(plugin.globals)
                self.filters.update(plugin.filters)


Template.environment_class = BettyEnvironment


class Jinja2Renderer(Renderer):
    def __init__(self, environment: Environment, configuration: Configuration):
        self._environment = environment
        self._configuration = configuration

    async def render_file(self, file_path: str) -> None:
        if not file_path.endswith('.j2'):
            return
        file_destination_path = file_path[:-3]
        data = {}
        if file_destination_path.startswith(self._configuration.www_directory_path):
            resource = '/'.join(Path(file_destination_path[len(self._configuration.www_directory_path):].strip(os.sep)).parts)
            if self._configuration.multilingual:
                resource_parts = resource.lstrip('/').split('/')
                if resource_parts[0] in map(lambda x: x.alias, self._configuration.locales.values()):
                    resource = '/'.join(resource_parts[1:])
            data['page_resource'] = resource
        root_path = rootname(file_path)
        template_name = '/'.join(Path(relpath(file_path, root_path)).parts)
        template = FileSystemLoader(root_path).load(self._environment, template_name, self._environment.globals)
        with open(file_destination_path, 'w') as f:
            f.write(await template.render_async(data))
        os.remove(file_path)

    async def render_tree(self, tree_path: str) -> None:
        await asyncio.gather(
            *[self.render_file(file_path) async for file_path in iterfiles(tree_path) if file_path.endswith('.j2')],
        )


@contextfilter
def _filter_url(context: Context, resource: Any, media_type: Optional[str] = None, locale: Optional[str] = None, **kwargs) -> str:
    media_type = 'text/html' if media_type is None else media_type
    locale = locale if locale else resolve_or_missing(context, 'locale')
    return context.environment.site.localized_url_generator.generate(resource, media_type, locale=locale, **kwargs)


@contextfilter
def _filter_json(context: Context, data: Any, indent: Optional[int] = None) -> str:
    """
    Converts a value to a JSON string.
    """
    return stdjson.dumps(data, indent=indent,
                         cls=JSONEncoder.get_factory(context.environment.site, resolve_or_missing(context, 'locale')))


@contextfilter
def _filter_tojson(context: Context, data: Any, indent: Optional[int] = None) -> str:
    """
    Converts a value to a JSON string safe for use in an HTML document.

    This mimics Jinja2's built-in JSON filter, but uses Betty's own JSON encoder.
    """
    return htmlsafe_json_dumps(data, indent=indent, dumper=lambda *args, **kwargs: _filter_json(context, *args, **kwargs))


async def _filter_flatten(items: Union[Iterable, AsyncIterable]) -> Iterable:
    async for item in _asynciter(items):
        async for child in _asynciter(item):
            yield child


def _filter_walk(item: Any, attribute_name: str) -> Iterable[Any]:
    return walk(item, attribute_name)


_paragraph_re = re.compile(r'(?:\r\n|\r|\n){2,}')


@evalcontextfilter
def _filter_paragraphs(eval_ctx: EvalContext, text: str) -> Union[str, Markup]:
    """Converts newlines to <p> and <br> tags.

    Taken from http://jinja.pocoo.org/docs/2.10/api/#custom-filters."""
    result = u'\n\n'.join(u'<p>%s</p>' % p.replace('\n', Markup('<br>\n'))
                          for p in _paragraph_re.split(escape(text)))
    if eval_ctx.autoescape:
        result = Markup(result)
    return result


def _filter_format_degrees(degrees: int) -> str:
    arcminutes = units.arcminutes(degrees=degrees - int(degrees))
    arcseconds = units.arcseconds(arcminutes=arcminutes - int(arcminutes))
    format_dict = dict(
        deg='°',
        arcmin="'",
        arcsec='"',
        degrees=degrees,
        minutes=round(abs(arcminutes)),
        seconds=round(abs(arcseconds))
    )
    return DEGREES_FORMAT % format_dict


async def _filter_unique(items: Iterable) -> Iterator:
    seen = []
    async for item in _asynciter(items):
        if item not in seen:
            yield item
            seen.append(item)


@contextfilter
async def _filter_map(*args, **kwargs):
    """
    Maps an iterable's values.

    This mimics Jinja2's built-in map filter, but allows macros as callbacks.
    """
    if len(args) == 3 and isinstance(args[2], Macro):
        seq = args[1]
        func = args[2]
    else:
        seq, func = prepare_map(args, kwargs)
    if seq:
        async for item in _asynciter(seq):
            yield await auto_await(func(item))


async def _filter_file(site: Site, file: File) -> str:
    file_directory_path = os.path.join(site.configuration.www_directory_path, 'file')

    destination_name = '%s.%s' % (file.id, file.extension)
    destination_public_path = '/file/%s' % destination_name

    with suppress(AcquiredError):
        site.locks.acquire((_filter_file, file))
        site.executor.submit(_do_filter_file, file.path, file_directory_path, destination_name)

    return destination_public_path


def _do_filter_file(file_path: str, destination_directory_path: str, destination_name: str) -> None:
    makedirs(destination_directory_path)
    destination_file_path = os.path.join(destination_directory_path, destination_name)
    link_or_copy(file_path, destination_file_path)


async def _filter_image(site: Site, file: File, width: Optional[int] = None, height: Optional[int] = None) -> str:
    if width is None and height is None:
        raise ValueError('At least the width or height must be given.')

    destination_name = '%s-' % file.id
    if width is None:
        destination_name += '-x%d' % height
    elif height is None:
        destination_name += '%dx-' % width
    else:
        destination_name += '%dx%d' % (width, height)

    file_directory_path = os.path.join(
        site.configuration.www_directory_path, 'file')

    if file.media_type:
        if file.media_type.type == 'image':
            task = _execute_filter_image_image
            destination_name += '.' + extension(file.path)
        elif file.media_type.type == 'application' and file.media_type.subtype == 'pdf':
            task = _execute_filter_image_application_pdf
            destination_name += '.' + 'jpg'
        else:
            raise ValueError('Cannot convert a file of media type "%s" to an image.' % file.media_type)
    else:
        raise ValueError('Cannot convert a file without a media type to an image.')

    with suppress(AcquiredError):
        site.locks.acquire((_filter_image, file, width, height))
        cache_directory_path = join(site.configuration.cache_directory_path, 'image')
        site.executor.submit(task, file.path, cache_directory_path, file_directory_path, destination_name, width, height)

    destination_public_path = '/file/%s' % destination_name

    return destination_public_path


def _execute_filter_image_image(file_path: str, *args, **kwargs) -> None:
    with warnings.catch_warnings():
        # Ignore warnings about decompression bombs, because we know where the files come from.
        warnings.simplefilter('ignore', category=DecompressionBombWarning)
        image = Image.open(file_path)
    _execute_filter_image(image, file_path, *args, **kwargs)


def _execute_filter_image_application_pdf(file_path: str, *args, **kwargs) -> None:
    with warnings.catch_warnings():
        # Ignore warnings about decompression bombs, because we know where the files come from.
        warnings.simplefilter('ignore', category=DecompressionBombWarning)
        image = pdf2image.convert_from_path(file_path, fmt='jpeg')[0]
    _execute_filter_image(image, file_path, *args, **kwargs)


def _execute_filter_image(image: Image, file_path: str, cache_directory_path: str, destination_directory_path: str, destination_name: str, width: int, height: int) -> None:
    makedirs(destination_directory_path)
    cache_file_path = join(cache_directory_path, '%s-%s' % (hashfile(file_path), destination_name))
    destination_file_path = join(destination_directory_path, destination_name)

    try:
        link_or_copy(cache_file_path, destination_file_path)
    except FileNotFoundError:
        makedirs(cache_directory_path)
        with image:
            if width is not None:
                width = min(width, image.width)
            if height is not None:
                height = min(height, image.height)

            if width is None:
                size = height
                convert = resizeimage.resize_height
            elif height is None:
                size = width
                convert = resizeimage.resize_width
            else:
                size = (width, height)
                convert = resizeimage.resize_cover
            convert(image, size).save(cache_file_path)
        makedirs(destination_directory_path)
        link_or_copy(cache_file_path, destination_file_path)


@contextfilter
def _filter_negotiate_localizeds(context: Context, localizeds: Iterable[Localized]) -> Optional[Localized]:
    locale = resolve_or_missing(context, 'locale')
    return negotiate_localizeds(locale, list(localizeds))


@contextfilter
def _filter_sort_localizeds(context: Context, localizeds: Iterable[Localized], localized_attribute: str, sort_attribute: str) -> Iterable[Localized]:
    locale = resolve_or_missing(context, 'locale')
    get_localized_attr = make_attrgetter(
        context.environment, localized_attribute)
    get_sort_attr = make_attrgetter(context.environment, sort_attribute)

    def _get_sort_key(x):
        return get_sort_attr(negotiate_localizeds(locale, get_localized_attr(x)))

    return sorted(localizeds, key=_get_sort_key)


@contextfilter
def _filter_select_localizeds(context: Context, localizeds: Iterable[Localized], include_unspecified: bool = False) -> Iterable[Localized]:
    locale = resolve_or_missing(context, 'locale')
    for localized in localizeds:
        if include_unspecified and localized.locale in {None, 'mis', 'mul', 'und', 'zxx'}:
            yield localized
        if localized.locale is not None and negotiate_locale(locale, [localized.locale]) is not None:
            yield localized


@contextfilter
def _filter_negotiate_dateds(context: Context, dateds: Iterable[Dated], date: Optional[Datey]) -> Optional[Dated]:
    with suppress(StopIteration):
        return next(_filter_select_dateds(context, dateds, date))


@contextfilter
def _filter_select_dateds(context: Context, dateds: Iterable[Dated], date: Optional[Datey]) -> Iterator[Dated]:
    if date is None:
        date = resolve_or_missing(context, 'today')
    return filter(lambda dated: dated.date is None or dated.date.comparable and dated.date in date, dateds)


@contextfilter
def _filter_format_date(context: Context, date: Datey) -> str:
    locale = resolve_or_missing(context, 'locale')
    return format_datey(date, locale)


async def _asynciter(items: Union[Iterable, AsyncIterable]) -> AsyncIterable:
    if hasattr(items, '__aiter__'):
        async for item in items:
            yield item
        return
    for item in items:
        yield item



import gettext
from tempfile import TemporaryDirectory
from unittest.mock import patch, ANY, Mock

from betty.ancestry import Ancestry, Person, File, Source, Citation, PersonName, Presence, Event, IdentifiableEvent, \
    IdentifiableSource, IdentifiableCitation, Birth, Subject, HasCitations
from betty.config import Configuration
from betty.asyncio import sync
from betty.parse import parse
from betty.plugin.anonymizer import Anonymizer, anonymize, anonymize_person, anonymize_event, anonymize_file, \
    anonymize_citation, anonymize_source, AnonymousSource, AnonymousCitation
from betty.site import Site
from betty.tests import TestCase


class AnonymousSourceTest(TestCase):
    def test_name(self):
        self.assertIsInstance(AnonymousSource().name, str)

    def test_replace(self):
        citations = [Citation(Source())]
        contains = [Source()]
        files = [Mock(File)]
        sut = AnonymousSource()
        other = AnonymousSource()
        other.citations = citations
        other.contains = contains
        other.files = files
        sut.replace(other)
        self.assertEquals(citations, list(sut.citations))
        self.assertEquals(contains, list(sut.contains))
        self.assertEquals(files, list(sut.files))


class AnonymousCitationTest(TestCase):
    def test_location(self):
        source = Mock(Source)
        self.assertIsInstance(AnonymousCitation(source).location, str)

    def test_replace(self):
        facts = [HasCitations()]
        files = [File('F1', __file__)]
        source = Mock(Source)
        sut = AnonymousCitation(source)
        other = AnonymousCitation(source)
        other.facts = facts
        other.files = files
        sut.replace(other)
        self.assertEquals(facts, list(sut.facts))
        self.assertEquals(files, list(sut.files))


class AnonymizeTest(TestCase):
    def setUp(self) -> None:
        gettext.NullTranslations().install()

    @patch('betty.plugin.anonymizer.anonymize_person')
    def test_with_public_person_should_not_anonymize(self, m_anonymize_person) -> None:
        person = Person('P0')
        person.private = False
        ancestry = Ancestry()
        ancestry.people[person.id] = person
        anonymize(ancestry)
        m_anonymize_person.assert_not_called()

    @patch('betty.plugin.anonymizer.anonymize_person')
    def test_with_private_person_should_anonymize(self, m_anonymize_person) -> None:
        person = Person('P0')
        person.private = True
        ancestry = Ancestry()
        ancestry.people[person.id] = person
        anonymize(ancestry)
        m_anonymize_person.assert_called_once_with(person)

    @patch('betty.plugin.anonymizer.anonymize_event')
    def test_with_public_event_should_not_anonymize(self, m_anonymize_event) -> None:
        event = IdentifiableEvent('E0', Birth())
        event.private = False
        ancestry = Ancestry()
        ancestry.events[event.id] = event
        anonymize(ancestry)
        m_anonymize_event.assert_not_called()

    @patch('betty.plugin.anonymizer.anonymize_event')
    def test_with_private_event_should_anonymize(self, m_anonymize_event) -> None:
        event = IdentifiableEvent('E0', Birth())
        event.private = True
        ancestry = Ancestry()
        ancestry.events[event.id] = event
        anonymize(ancestry)
        m_anonymize_event.assert_called_once_with(event)

    @patch('betty.plugin.anonymizer.anonymize_file')
    def test_with_public_file_should_not_anonymize(self, m_anonymize_file) -> None:
        file = File('F0', __file__)
        file.private = False
        ancestry = Ancestry()
        ancestry.files[file.id] = file
        anonymize(ancestry)
        m_anonymize_file.assert_not_called()

    @patch('betty.plugin.anonymizer.anonymize_file')
    def test_with_private_file_should_anonymize(self, m_anonymize_file) -> None:
        file = File('F0', __file__)
        file.private = True
        ancestry = Ancestry()
        ancestry.files[file.id] = file
        anonymize(ancestry)
        m_anonymize_file.assert_called_once_with(file)

    @patch('betty.plugin.anonymizer.anonymize_source')
    def test_with_public_source_should_not_anonymize(self, m_anonymize_source) -> None:
        source = IdentifiableSource('S0', 'The Source')
        source.private = False
        ancestry = Ancestry()
        ancestry.sources[source.id] = source
        anonymize(ancestry)
        m_anonymize_source.assert_not_called()

    @patch('betty.plugin.anonymizer.anonymize_source')
    def test_with_private_source_should_anonymize(self, m_anonymize_source) -> None:
        source = IdentifiableSource('S0', 'The Source')
        source.private = True
        ancestry = Ancestry()
        ancestry.sources[source.id] = source
        anonymize(ancestry)
        m_anonymize_source.assert_called_once_with(source, ANY)

    @patch('betty.plugin.anonymizer.anonymize_citation')
    def test_with_public_citation_should_not_anonymize(self, m_anonymize_citation) -> None:
        source = Source('The Source')
        citation = IdentifiableCitation('C0', source)
        citation.private = False
        ancestry = Ancestry()
        ancestry.citations[citation.id] = citation
        anonymize(ancestry)
        m_anonymize_citation.assert_not_called()

    @patch('betty.plugin.anonymizer.anonymize_citation')
    def test_with_private_citation_should_anonymize(self, m_anonymize_citation) -> None:
        source = Source('The Source')
        citation = IdentifiableCitation('C0', source)
        citation.private = True
        ancestry = Ancestry()
        ancestry.citations[citation.id] = citation
        anonymize(ancestry)
        m_anonymize_citation.assert_called_once_with(citation, ANY)


class AnonymizePersonTest(TestCase):
    def test_should_remove_citations(self) -> None:
        person = Person('P0')
        source = Source('The Source')
        citation = Citation(source)
        person.citations.append(citation)
        anonymize_person(person)
        self.assertEquals(0, len(person.citations))

    def test_should_remove_files(self) -> None:
        person = Person('P0')
        person.files.append(File('F0', __file__))
        anonymize_person(person)
        self.assertEquals(0, len(person.files))

    def test_should_remove_names(self) -> None:
        person = Person('P0')
        name = PersonName('Jane', 'Dough')
        source = Source('The Source')
        citation = Citation(source)
        name.citations.append(citation)
        person.names.append(name)
        anonymize_person(person)
        self.assertEquals(0, len(person.names))
        self.assertEquals(0, len(citation.facts))

    def test_should_remove_presences(self) -> None:
        person = Person('P0')
        event = Event(Birth())
        Presence(person, Subject(), event)
        anonymize_person(person)
        self.assertEquals(0, len(person.presences))
        self.assertEquals(0, len(event.presences))

    def test_should_remove_parents_without_public_descendants(self) -> None:
        person = Person('P0')
        person.private = True
        child = Person('P1')
        child.private = True
        person.children.append(child)
        parent = Person('P2')
        parent.private = True
        person.parents.append(parent)

        anonymize_person(person)
        self.assertCountEqual([], person.parents)

    def test_should_not_remove_parents_with_public_descendants(self) -> None:
        person = Person('P0')
        person.private = True
        child = Person('P1')
        child.private = False
        person.children.append(child)
        parent = Person('P2')
        parent.private = True
        person.parents.append(parent)

        anonymize_person(person)
        self.assertCountEqual([parent], person.parents)


class AnonymizeEventTest(TestCase):
    def test_should_remove_citations(self) -> None:
        event = Event(Birth())
        source = Source('The Source')
        citation = Citation(source)
        event.citations.append(citation)
        anonymize_event(event)
        self.assertEquals(0, len(event.citations))

    def test_should_remove_files(self) -> None:
        event = Event(Birth())
        event.files.append(File('F0', __file__))
        anonymize_event(event)
        self.assertEquals(0, len(event.files))

    def test_should_remove_presences(self) -> None:
        event = Event(Birth())
        person = Person('P1')
        Presence(person, Subject(), event)
        anonymize_event(event)
        self.assertEquals(0, len(event.presences))


class AnonymizeFileTest(TestCase):
    def test_should_remove_resources(self) -> None:
        file = File('F0', __file__)
        file.resources.append(Person('P0'))
        anonymize_file(file)
        self.assertEquals(0, len(file.resources))


class AnonymizeSourceTest(TestCase):
    def setUp(self) -> None:
        gettext.NullTranslations().install()

    def test_should_remove_citations(self) -> None:
        source = IdentifiableSource('S0', 'The Source')
        citation = Citation(source)
        source.citations.append(citation)
        anonymous_source = AnonymousSource()
        anonymize_source(source, anonymous_source)
        self.assertEquals(0, len(source.citations))
        self.assertIn(citation, anonymous_source.citations)

    def test_should_remove_contained_by(self) -> None:
        source = IdentifiableSource('S0', 'The Source')
        contained_by = Source('The Source')
        source.contained_by = contained_by
        anonymous_source = AnonymousSource()
        anonymize_source(source, anonymous_source)
        self.assertIsNone(source.contained_by)

    def test_should_remove_contains(self) -> None:
        source = IdentifiableSource('S0', 'The Source')
        contains = Source('The Source')
        source.contains.append(contains)
        anonymous_source = AnonymousSource()
        anonymize_source(source, anonymous_source)
        self.assertEquals(0, len(source.contains))
        self.assertIn(contains, anonymous_source.contains)

    def test_should_remove_files(self) -> None:
        source = IdentifiableSource('S0', 'The Source')
        file = File('F0', __file__)
        source.files.append(file)
        anonymous_source = AnonymousSource()
        anonymize_source(source, anonymous_source)
        self.assertEquals(0, len(source.files))
        self.assertIn(file, anonymous_source.files)


class AnonymizeCitationTest(TestCase):
    def setUp(self) -> None:
        gettext.NullTranslations().install()

    def test_should_remove_facts(self) -> None:
        source = Source('The Source')
        citation = IdentifiableCitation('C0', source)
        fact = PersonName('Jane')
        citation.facts.append(fact)
        anonymous_source = AnonymousSource()
        anonymous_citation = AnonymousCitation(anonymous_source)
        anonymize_citation(citation, anonymous_citation)
        self.assertEquals(0, len(citation.facts))
        self.assertIn(fact, anonymous_citation.facts)

    def test_should_remove_files(self) -> None:
        source = Source('The Source')
        citation = IdentifiableCitation('C0', source)
        file = File('F0', __file__)
        citation.files.append(file)
        anonymous_source = AnonymousSource()
        anonymous_citation = AnonymousCitation(anonymous_source)
        anonymize_citation(citation, anonymous_citation)
        self.assertEquals(0, len(citation.files))
        self.assertIn(file, anonymous_citation.files)

    def test_should_remove_source(self) -> None:
        source = Source('The Source')
        citation = IdentifiableCitation('C0', source)
        anonymous_source = AnonymousSource()
        anonymous_citation = AnonymousCitation(anonymous_source)
        anonymize_citation(citation, anonymous_citation)
        self.assertIsNone(citation.source)


class AnonymizerTest(TestCase):
    @sync
    async def test_post_parse(self) -> None:
        person = Person('P0')
        person.private = True
        person.names.append(PersonName('Jane', 'Dough'))
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.plugins[Anonymizer] = None
            async with Site(configuration) as site:
                site.ancestry.people[person.id] = person
                await parse(site)
        self.assertEquals(0, len(person.names))



from os import path
from os.path import splitext
from typing import Optional


def rootname(source_path: str) -> str:
    root = source_path
    while True:
        possible_root = path.dirname(root)
        if possible_root == root:
            return root
        root = possible_root


def extension(path: str) -> Optional[str]:
    extension = splitext(path)[1][1:]
    return extension if extension else None



from os import path
from tempfile import TemporaryDirectory
from time import sleep

import requests

from betty.asyncio import sync
from betty.serve import BuiltinServer
from betty.tests import TestCase


class BuiltinServerTest(TestCase):
    @sync
    async def test(self):
        content = 'Hello, and welcome to my site!'
        with TemporaryDirectory() as www_directory_path:
            with open(path.join(www_directory_path, 'index.html'), 'w') as f:
                f.write(content)
            async with BuiltinServer(www_directory_path) as server:
                # Wait for the server to start.
                sleep(1)
                response = requests.get(server.public_url)
                self.assertEquals(200, response.status_code)
                self.assertEquals(content, response.content.decode('utf-8'))



from os import path
from tempfile import TemporaryDirectory
from unittest.mock import patch

from betty.os import link_or_copy
from betty.tests import TestCase


class LinkOrCopyTest(TestCase):
    def test(self):
        with TemporaryDirectory() as working_directory_path:
            content = 'I will say zis only once.'
            source_path = path.join(working_directory_path, 'source')
            destination_path = path.join(working_directory_path, 'destination')
            with open(source_path, 'a') as f:
                f.write(content)
            link_or_copy(source_path, destination_path)
            with open(destination_path) as f:
                self.assertEqual(content, f.read())

    @patch('os.link')
    def test_with_os_error(self, m_link):
        m_link.side_effect = OSError
        with TemporaryDirectory() as working_directory_path:
            content = 'I will say zis only once.'
            source_path = path.join(working_directory_path, 'source')
            destination_path = path.join(working_directory_path, 'destination')
            with open(source_path, 'a') as f:
                f.write(content)
            link_or_copy(source_path, destination_path)
            with open(destination_path) as f:
                self.assertEqual(content, f.read())



from typing import Set


class HtmlProvider:
    @property
    def public_css_paths(self) -> Set[str]:
        return set()

    @property
    def public_js_paths(self) -> Set[str]:
        return set()



from tempfile import TemporaryDirectory

from parameterized import parameterized

from betty.ancestry import Person, Place, PlaceName, PersonName, File
from betty.config import Configuration, LocaleConfiguration
from betty.asyncio import sync
from betty.search import Index
from betty.site import Site
from betty.tests import TestCase


class IndexTest(TestCase):
    @sync
    async def test_empty(self):
        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.locales['en-US'] = LocaleConfiguration('en-US', 'en')
            configuration.locales['nl-NL'] = LocaleConfiguration('nl-NL', 'nl')
            async with Site(configuration) as site:
                indexed = [item for item in await Index(site).build()]

        self.assertEquals([], indexed)

    @sync
    async def test_person_without_names(self):
        person_id = 'P1'
        person = Person(person_id)

        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.locales['en-US'] = LocaleConfiguration('en-US', 'en')
            configuration.locales['nl-NL'] = LocaleConfiguration('nl-NL', 'nl')
            async with Site(configuration) as site:
                site.ancestry.people[person_id] = person
                indexed = [item for item in await Index(site).build()]

        self.assertEquals([], indexed)

    @sync
    async def test_private_person(self):
        person_id = 'P1'
        individual_name = 'Jane'
        person = Person(person_id)
        person.names.append(PersonName(individual_name))
        person.private = True

        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.locales['en-US'] = LocaleConfiguration('en-US', 'en')
            configuration.locales['nl-NL'] = LocaleConfiguration('nl-NL', 'nl')
            async with Site(configuration) as site:
                site.ancestry.people[person_id] = person
                indexed = [item for item in await Index(site).build()]

        self.assertEquals([], indexed)

    @parameterized.expand([
        ('/nl/person/P1/index.html', 'nl-NL'),
        ('/en/person/P1/index.html', 'en-US'),
    ])
    @sync
    async def test_person_with_individual_name(self, expected: str, locale: str):
        person_id = 'P1'
        individual_name = 'Jane'
        person = Person(person_id)
        person.names.append(PersonName(individual_name))

        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.locales['en-US'] = LocaleConfiguration('en-US', 'en')
            configuration.locales['nl-NL'] = LocaleConfiguration('nl-NL', 'nl')
            async with Site(configuration).with_locale(locale) as site:
                site.ancestry.people[person_id] = person
                indexed = [item for item in await Index(site).build()]

        self.assertEquals('jane', indexed[0]['text'])
        self.assertIn(expected, indexed[0]['result'])

    @parameterized.expand([
        ('/nl/person/P1/index.html', 'nl-NL'),
        ('/en/person/P1/index.html', 'en-US'),
    ])
    @sync
    async def test_person_with_affiliation_name(self, expected: str, locale: str):
        person_id = 'P1'
        affiliation_name = 'Doughnut'
        person = Person(person_id)
        person.names.append(PersonName(None, affiliation_name))

        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.locales['en-US'] = LocaleConfiguration('en-US', 'en')
            configuration.locales['nl-NL'] = LocaleConfiguration('nl-NL', 'nl')
            async with Site(configuration).with_locale(locale) as site:
                site.ancestry.people[person_id] = person
                indexed = [item for item in await Index(site).build()]

        self.assertEquals('doughnut', indexed[0]['text'])
        self.assertIn(expected, indexed[0]['result'])

    @parameterized.expand([
        ('/nl/person/P1/index.html', 'nl-NL'),
        ('/en/person/P1/index.html', 'en-US'),
    ])
    @sync
    async def test_person_with_individual_and_affiliation_names(self, expected: str, locale: str):
        person_id = 'P1'
        individual_name = 'Jane'
        affiliation_name = 'Doughnut'
        person = Person(person_id)
        person.names.append(PersonName(individual_name, affiliation_name))

        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.locales['en-US'] = LocaleConfiguration('en-US', 'en')
            configuration.locales['nl-NL'] = LocaleConfiguration('nl-NL', 'nl')
            async with Site(configuration).with_locale(locale) as site:
                site.ancestry.people[person_id] = person
                indexed = [item for item in await Index(site).build()]

        self.assertEquals('jane doughnut', indexed[0]['text'])
        self.assertIn(expected, indexed[0]['result'])

    @parameterized.expand([
        ('/nl/place/P1/index.html', 'nl-NL'),
        ('/en/place/P1/index.html', 'en-US'),
    ])
    @sync
    async def test_place(self, expected: str, locale: str):
        place_id = 'P1'
        place = Place(place_id, [PlaceName('Netherlands', 'en'), PlaceName('Nederland', 'nl')])

        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.locales['en-US'] = LocaleConfiguration('en-US', 'en')
            configuration.locales['nl-NL'] = LocaleConfiguration('nl-NL', 'nl')
            async with Site(configuration).with_locale(locale) as site:
                site.ancestry.places[place_id] = place
                indexed = [item for item in await Index(site).build()]

        self.assertEquals('netherlands nederland', indexed[0]['text'])
        self.assertIn(expected, indexed[0]['result'])

    @sync
    async def test_file_without_description(self):
        file_id = 'F1'
        file = File(file_id, __file__)

        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.locales['en-US'] = LocaleConfiguration('en-US', 'en')
            configuration.locales['nl-NL'] = LocaleConfiguration('nl-NL', 'nl')
            async with Site(configuration) as site:
                site.ancestry.files[file_id] = file
                indexed = [item for item in await Index(site).build()]

        self.assertEquals([], indexed)

    @parameterized.expand([
        ('/nl/file/F1/index.html', 'nl-NL'),
        ('/en/file/F1/index.html', 'en-US'),
    ])
    @sync
    async def test_file(self, expected: str, locale: str):
        file_id = 'F1'
        file = File(file_id, __file__)
        file.description = '"file" is Dutch for "traffic jam"'

        with TemporaryDirectory() as output_directory_path:
            configuration = Configuration(
                output_directory_path, 'https://example.com')
            configuration.locales['en-US'] = LocaleConfiguration('en-US', 'en')
            configuration.locales['nl-NL'] = LocaleConfiguration('nl-NL', 'nl')
            async with Site(configuration).with_locale(locale) as site:
                site.ancestry.files[file_id] = file
                indexed = [item for item in await Index(site).build()]

        self.assertEquals('"file" is dutch for "traffic jam"', indexed[0]['text'])
        self.assertIn(expected, indexed[0]['result'])



from typing import Any, Iterable


def walk(item: Any, attribute_name: str) -> Iterable[Any]:
    child = getattr(item, attribute_name)

    # If the child has the requested attribute, yield it,
    if hasattr(child, attribute_name):
        yield child
        yield from walk(child, attribute_name)

    # Otherwise loop over the children and yield their attributes.
    try:
        child = iter(child)
    except TypeError:
        return
    for child in child:
        yield child
        yield from walk(child, attribute_name)



from collections import defaultdict
from typing import Set, Type, Any

from betty.ancestry import Ancestry, Place, File, IdentifiableEvent, IdentifiableSource, IdentifiableCitation, Person
from betty.graph import Graph, tsort
from betty.parse import PostParser
from betty.plugin import Plugin, NO_CONFIGURATION
from betty.plugin.anonymizer import Anonymizer
from betty.site import Site


def clean(ancestry: Ancestry) -> None:
    _clean_people(ancestry)
    _clean_events(ancestry)
    _clean_places(ancestry)
    _clean_files(ancestry)
    _clean_citations(ancestry)
    _clean_sources(ancestry)


def _clean_events(ancestry: Ancestry) -> None:
    for event in list(ancestry.events.values()):
        _clean_event(ancestry, event)


def _clean_event(ancestry: Ancestry, event: IdentifiableEvent) -> None:
    if len(event.presences) > 0:
        return

    del event.presences
    del event.place
    del event.citations
    del event.files
    del ancestry.events[event.id]


def _clean_places(ancestry: Ancestry) -> None:
    places = list(ancestry.places.values())

    def _extend_place_graph(graph: Graph, enclosing_place: Place) -> None:
        enclosures = enclosing_place.encloses
        # Ensure each place appears in the graph, even if they're anonymous.
        graph.setdefault(enclosing_place, set())
        for enclosure in enclosures:
            enclosed_place = enclosure.encloses
            seen_enclosed_place = enclosed_place in graph
            graph[enclosed_place].add(enclosing_place)
            if not seen_enclosed_place:
                _extend_place_graph(graph, enclosed_place)

    places_graph = defaultdict(set)
    for place in places:
        _extend_place_graph(places_graph, place)

    for place in tsort(places_graph):
        _clean_place(ancestry, place)


def _clean_place(ancestry: Ancestry, place: Place) -> None:
    if len(place.events) > 0:
        return

    if len(place.encloses) > 0:
        return

    del place.enclosed_by
    del ancestry.places[place.id]


def _clean_people(ancestry: Ancestry) -> None:
    for person in list(ancestry.people.values()):
        _clean_person(ancestry, person)


def _clean_person(ancestry: Ancestry, person: Person) -> None:
    if not person.private:
        return

    if len(person.children) > 0:
        return

    del ancestry.people[person.id]


def _clean_files(ancestry: Ancestry) -> None:
    for file in list(ancestry.files.values()):
        _clean_file(ancestry, file)


def _clean_file(ancestry: Ancestry, file: File) -> None:
    if len(file.resources) > 0:
        return

    del ancestry.files[file.id]


def _clean_sources(ancestry: Ancestry) -> None:
    for source in list(ancestry.sources.values()):
        _clean_source(ancestry, source)


def _clean_source(ancestry: Ancestry, source: IdentifiableSource) -> None:
    if len(source.citations) > 0:
        return

    if source.contained_by is not None:
        return

    if len(source.contains) > 0:
        return

    if len(source.files) > 0:
        return

    del ancestry.sources[source.id]


def _clean_citations(ancestry: Ancestry) -> None:
    for citation in list(ancestry.citations.values()):
        _clean_citation(ancestry, citation)


def _clean_citation(ancestry: Ancestry, citation: IdentifiableCitation) -> None:
    if len(citation.facts) > 0:
        return

    if len(citation.files) > 0:
        return

    del citation.source
    del ancestry.citations[citation.id]


class Cleaner(Plugin, PostParser):
    def __init__(self, ancestry: Ancestry):
        self._ancestry = ancestry

    @classmethod
    def for_site(cls, site: Site, configuration: Any = NO_CONFIGURATION):
        return cls(site.ancestry)

    @classmethod
    def comes_after(cls) -> Set[Type[Plugin]]:
        return {Anonymizer}

    async def post_parse(self) -> None:
        clean(self._ancestry)



from tempfile import TemporaryDirectory
from time import sleep
from typing import Tuple, Optional
from unittest.mock import patch, call

from betty.media_type import MediaType
from betty.tests import TestCase

try:
    from unittest.mock import AsyncMock
except ImportError:
    from mock.mock import AsyncMock

import aiohttp
from aioresponses import aioresponses
from parameterized import parameterized

from betty.ancestry import Source, IdentifiableCitation, IdentifiableSource, Link
from betty.config import Configuration, LocaleConfiguration
from betty.asyncio import sync
from betty.parse import parse
from betty.plugin.wikipedia import Entry, Retriever, NotAnEntryError, parse_url, RetrievalError, _Populator, Wikipedia
from betty.site import Site


class ParseUrlTest(TestCase):
    @parameterized.expand([
        (('en', 'Amsterdam'), 'http://en.wikipedia.org/wiki/Amsterdam',),
        (('nl', 'Amsterdam'), 'https://nl.wikipedia.org/wiki/Amsterdam',),
        (('en', 'Amsterdam'), 'http://en.wikipedia.org/wiki/Amsterdam',),
        (('en', 'Amsterdam'), 'https://en.wikipedia.org/wiki/Amsterdam/',),
        (('en', 'Amsterdam'), 'https://en.wikipedia.org/wiki/Amsterdam/some-path',),
        (('en', 'Amsterdam'), 'https://en.wikipedia.org/wiki/Amsterdam?some=query',),
        (('en', 'Amsterdam'), 'https://en.wikipedia.org/wiki/Amsterdam#some-fragment',),
    ])
    def test_should_return(self, expected: Tuple[str, str], url: str) -> None:
        self.assertEquals(expected, parse_url(url))

    @parameterized.expand([
        ('',),
        ('ftp://en.wikipedia.org/wiki/Amsterdam',),
        ('https://en.wikipedia.org/w/index.php?title=Amsterdam&action=edit',),
    ])
    def test_should_error(self, url: str) -> None:
        with self.assertRaises(NotAnEntryError):
            parse_url(url)


class EntryTest(TestCase):
    def test_url(self) -> None:
        sut = Entry('nl', 'Amsterdam', 'Title for Amsterdam', 'Content for Amsterdam')
        self.assertEquals('https://nl.wikipedia.org/wiki/Amsterdam', sut.url)

    def test_title(self) -> None:
        title = 'Title for Amsterdam'
        sut = Entry('nl', 'Amsterdam', title, 'Content for Amsterdam')
        self.assertEquals(title, sut.title)

    def test_content(self) -> None:
        content = 'Content for Amsterdam'
        sut = Entry('nl', 'Amsterdam', 'Title for Amsterdam', content)
        self.assertEquals(content, sut.content)


class RetrieverTest(TestCase):
    @parameterized.expand([
        ({}, {},),
        ({
            'nl': 'Amsterdam',
            'uk': 'Амстердам',
        }, {
            'langlinks': [
                {
                    'lang': 'nl',
                    'title': 'Amsterdam',
                },
                {
                    'lang': 'uk',
                    'title': 'Амстердам',
                },
            ],
        },),
    ])
    @aioresponses()
    @patch('sys.stderr')
    @sync
    async def test_get_translations_should_return(self, expected, response_pages_json, m_stderr, m_aioresponses) -> None:
        entry_language = 'en'
        entry_name = 'Amsterdam'
        api_url = 'https://%s.wikipedia.org/w/api.php?action=query&titles=%s&prop=langlinks&lllimit=500&format=json&formatversion=2' % (entry_language, entry_name)
        api_response_body = {
            'query': {
                'pages': [response_pages_json],
            },
        }
        m_aioresponses.get(api_url, payload=api_response_body)
        with TemporaryDirectory() as cache_directory_path:
            async with aiohttp.ClientSession() as session:
                translations = await Retriever(session, cache_directory_path).get_translations(entry_language, entry_name)
        self.assertEqual(expected, translations)

    @aioresponses()
    @patch('sys.stderr')
    @sync
    async def test_get_translations_with_client_error_should_raise_retrieval_error(self, m_aioresponses, m_stderr) -> None:
        entry_language = 'en'
        entry_name = 'Amsterdam'
        api_url = 'https://%s.wikipedia.org/w/api.php?action=query&titles=%s&prop=langlinks&lllimit=500&format=json&formatversion=2' % (entry_language, entry_name)
        m_aioresponses.get(api_url, exception=aiohttp.ClientError())
        with TemporaryDirectory() as cache_directory_path:
            with self.assertRaises(RetrievalError):
                async with aiohttp.ClientSession() as session:
                    await Retriever(session, cache_directory_path).get_translations(entry_language, entry_name)

    @aioresponses()
    @patch('sys.stderr')
    @sync
    async def test_get_translations_with_invalid_json_response_should_raise_retrieval_error(self, m_aioresponses, m_stderr) -> None:
        entry_language = 'en'
        entry_name = 'Amsterdam'
        api_url = 'https://%s.wikipedia.org/w/api.php?action=query&titles=%s&prop=langlinks&lllimit=500&format=json&formatversion=2' % (entry_language, entry_name)
        m_aioresponses.get(api_url, body='{Haha Im not rly JSON}')
        with TemporaryDirectory() as cache_directory_path:
            with self.assertRaises(RetrievalError):
                async with aiohttp.ClientSession() as session:
                    await Retriever(session, cache_directory_path).get_translations(entry_language, entry_name)

    @parameterized.expand([
        ({},),
        ({
            'query': {}
        },),
        ({
            'query': {
                'pages': {}
            }
        },),
        ({
            'query': {
                'pages': []
            }
        },),
    ])
    @aioresponses()
    @patch('sys.stderr')
    @sync
    async def test_get_translations_with_unexpected_json_response_should_raise_retrieval_error(self, response_json, m_stderr, m_aioresponses) -> None:
        entry_language = 'en'
        entry_name = 'Amsterdam'
        api_url = 'https://%s.wikipedia.org/w/api.php?action=query&titles=%s&prop=langlinks&lllimit=500&format=json&formatversion=2' % (entry_language, entry_name)
        m_aioresponses.get(api_url, payload=response_json)
        with TemporaryDirectory() as cache_directory_path:
            with self.assertRaises(RetrievalError):
                async with aiohttp.ClientSession() as session:
                    await Retriever(session, cache_directory_path).get_translations(entry_language, entry_name)

    @aioresponses()
    @sync
    async def test_get_entry_should_return(self, m_aioresponses) -> None:
        entry_language = 'en'
        entry_name = 'Amsterdam'
        api_url = 'https://en.wikipedia.org/w/api.php?action=query&titles=Amsterdam&prop=extracts&exintro&format=json&formatversion=2'
        entry_url = 'https://en.wikipedia.org/wiki/Amsterdam'
        title = 'Amstelredam'
        extract_1 = 'De hoofdstad van Nederland.'
        extract_4 = 'Niet de hoofdstad van Holland.'
        api_response_body_1 = {
            'query': {
                'pages': [
                    {
                        'title': title,
                        'extract': extract_1,
                    },
                ],
            }
        }
        api_response_body_4 = {
            'query': {
                'pages': [
                    {
                        'title': title,
                        'extract': extract_4,
                    },
                ],
            }
        }
        m_aioresponses.get(api_url, payload=api_response_body_1)
        m_aioresponses.get(api_url, exception=aiohttp.ClientError())
        m_aioresponses.get(api_url, payload=api_response_body_4)
        with TemporaryDirectory() as cache_directory_path:
            async with aiohttp.ClientSession() as session:
                retriever = Retriever(session, cache_directory_path, 1)
                # The first retrieval should make a successful request and set the cache.
                entry_1 = await retriever.get_entry(entry_language, entry_name)
                # The second retrieval should hit the cache from the first request.
                entry_2 = await retriever.get_entry(entry_language, entry_name)
                # The third retrieval should result in a failed request, and hit the cache from the first request.
                sleep(2)
                entry_3 = await retriever.get_entry(entry_language, entry_name)
                # The fourth retrieval should make a successful request and set the cache again.
                entry_4 = await retriever.get_entry(entry_language, entry_name)
                # The fifth retrieval should hit the cache from the fourth request.
                entry_5 = await retriever.get_entry(entry_language, entry_name)
        for entry in [entry_1, entry_2, entry_3]:
            self.assertEquals(entry_url, entry.url)
            self.assertEquals(title, entry.title)
            self.assertEquals(extract_1, entry.content)
        for entry in [entry_4, entry_5]:
            self.assertEquals(entry_url, entry.url)
            self.assertEquals(title, entry.title)
            self.assertEquals(extract_4, entry.content)

    @aioresponses()
    @patch('sys.stderr')
    @sync
    async def test_get_entry_with_client_error_should_raise_retrieval_error(self, m_aioresponses, m_stderr) -> None:
        entry_language = 'en'
        entry_name = 'Amsterdam'
        api_url = 'https://en.wikipedia.org/w/api.php?action=query&titles=Amsterdam&prop=extracts&exintro&format=json&formatversion=2'
        m_aioresponses.get(api_url, exception=aiohttp.ClientError())
        with TemporaryDirectory() as cache_directory_path:
            async with aiohttp.ClientSession() as session:
                retriever = Retriever(session, cache_directory_path)
                with self.assertRaises(RetrievalError):
                    await retriever.get_entry(entry_language, entry_name)


class PopulatorTest(TestCase):
    @patch('betty.plugin.wikipedia.Retriever')
    @sync
    async def test_populate_link_should_convert_http_to_https(self, m_retriever) -> None:
        link = Link('http://en.wikipedia.org/wiki/Amsterdam')
        entry_language = 'nl'
        with TemporaryDirectory() as output_directory_path:
            with TemporaryDirectory() as cache_directory_path:
                configuration = Configuration(
                    output_directory_path, 'https://example.com')
                configuration.cache_directory_path = cache_directory_path
                async with Site(configuration) as site:
                    sut = _Populator(site, m_retriever)
                    await sut.populate_link(link, entry_language)
        self.assertEqual('https://en.wikipedia.org/wiki/Amsterdam', link.url)

    @parameterized.expand([
        (MediaType('text/plain'), MediaType('text/plain')),
        (MediaType('text/html'), MediaType('text/html')),
        (MediaType('text/html'), None),
    ])
    @patch('betty.plugin.wikipedia.Retriever')
    @sync
    async def test_populate_link_should_set_media_type(self, expected: MediaType, media_type: Optional[MediaType], m_retriever) -> None:
        link = Link('http://en.wikipedia.org/wiki/Amsterdam')
        link.media_type = media_type
        with TemporaryDirectory() as output_directory_path:
            with TemporaryDirectory() as cache_directory_path:
                configuration = Configuration(
                    output_directory_path, 'https://example.com')
                configuration.cache_directory_path = cache_directory_path
                async with Site(configuration) as site:
                    sut = _Populator(site, m_retriever)
                    await sut.populate_link(link, 'en')
        self.assertEqual(expected, link.media_type)

    @parameterized.expand([
        ('alternate', 'alternate'),
        ('external', 'external'),
        ('external', None),
    ])
    @patch('betty.plugin.wikipedia.Retriever')
    @sync
    async def test_populate_link_should_set_relationship(self, expected: str, relationship: Optional[str], m_retriever) -> None:
        link = Link('http://en.wikipedia.org/wiki/Amsterdam')
        link.relationship = relationship
        with TemporaryDirectory() as output_directory_path:
            with TemporaryDirectory() as cache_directory_path:
                configuration = Configuration(
                    output_directory_path, 'https://example.com')
                configuration.cache_directory_path = cache_directory_path
                async with Site(configuration) as site:
                    sut = _Populator(site, m_retriever)
                    await sut.populate_link(link, 'en')
        self.assertEqual(expected, link.relationship)

    @parameterized.expand([
        ('nl-NL', 'nl', 'nl-NL'),
        ('nl', 'nl', None),
        ('nl', 'en', 'nl'),
    ])
    @patch('betty.plugin.wikipedia.Retriever')
    @sync
    async def test_populate_link_should_set_locale(self, expected: str, entry_language: str, locale: Optional[str], m_retriever) -> None:
        link = Link('http://%s.wikipedia.org/wiki/Amsterdam' % entry_language)
        link.locale = locale
        with TemporaryDirectory() as output_directory_path:
            with TemporaryDirectory() as cache_directory_path:
                configuration = Configuration(
                    output_directory_path, 'https://example.com')
                configuration.cache_directory_path = cache_directory_path
                async with Site(configuration) as site:
                    sut = _Populator(site, m_retriever)
                    await sut.populate_link(link, entry_language)
        self.assertEqual(expected, link.locale)

    @parameterized.expand([
        ('This is the original description', 'This is the original description'),
        ('Read more on Wikipedia.', None),
    ])
    @patch('betty.plugin.wikipedia.Retriever')
    @sync
    async def test_populate_link_should_set_description(self, expected: str, description: str, m_retriever) -> None:
        link = Link('http://en.wikipedia.org/wiki/Amsterdam')
        link.description = description
        entry_language = 'en'
        with TemporaryDirectory() as output_directory_path:
            with TemporaryDirectory() as cache_directory_path:
                configuration = Configuration(
                    output_directory_path, 'https://example.com')
                configuration.cache_directory_path = cache_directory_path
                async with Site(configuration) as site:
                    sut = _Populator(site, m_retriever)
                    await sut.populate_link(link, entry_language)
        self.assertEqual(expected, link.description)

    @parameterized.expand([
        ('Amsterdam', 'Amsterdam'),
        ('The city of Amsterdam', None),
    ])
    @patch('betty.plugin.wikipedia.Retriever')
    @sync
    async def test_populate_link_should_set_label(self, expected: str, label: Optional[str], m_retriever) -> None:
        link = Link('http://en.wikipedia.org/wiki/Amsterdam')
        link.label = label
        entry = Entry('en', 'The_city_of_Amsterdam', 'The city of Amsterdam', 'Amsterdam, such a lovely place!')
        with TemporaryDirectory() as output_directory_path:
            with TemporaryDirectory() as cache_directory_path:
                configuration = Configuration(
                    output_directory_path, 'https://example.com')
                configuration.cache_directory_path = cache_directory_path
                async with Site(configuration) as site:
                    sut = _Populator(site, m_retriever)
                    await sut.populate_link(link, 'en', entry)
        self.assertEqual(expected, link.label)

    @patch('betty.plugin.wikipedia.Retriever')
    @sync
    async def test_populate_should_ignore_resource_without_link_support(self, m_retriever) -> None:
        source = Source('The Source')
        resource = IdentifiableCitation('the_citation', source)
        with TemporaryDirectory() as output_directory_path:
            with TemporaryDirectory() as cache_directory_path:
                configuration = Configuration(
                    output_directory_path, 'https://example.com')
                configuration.cache_directory_path = cache_directory_path
                async with Site(configuration) as site:
                    site.ancestry.citations[resource.id] = resource
                    sut = _Populator(site, m_retriever)
                    await sut.populate()

    @patch('betty.plugin.wikipedia.Retriever')
    @sync
    async def test_populate_should_ignore_resource_without_links(self, m_retriever) -> None:
        resource = IdentifiableSource('the_source', 'The Source')
        with TemporaryDirectory() as output_directory_path:
            with TemporaryDirectory() as cache_directory_path:
                configuration = Configuration(
                    output_directory_path, 'https://example.com')
                configuration.cache_directory_path = cache_directory_path
                async with Site(configuration) as site:
                    site.ancestry.sources[resource.id] = resource
                    sut = _Populator(site, m_retriever)
                    await sut.populate()
        self.assertSetEqual(set(), resource.links)

    @patch('betty.plugin.wikipedia.Retriever')
    @sync
    async def test_populate_should_ignore_non_wikipedia_links(self, m_retriever) -> None:
        link = Link('https://example.com')
        resource = IdentifiableSource('the_source', 'The Source')
        resource.links.add(link)
        with TemporaryDirectory() as output_directory_path:
            with TemporaryDirectory() as cache_directory_path:
                configuration = Configuration(
                    output_directory_path, 'https://example.com')
                configuration.cache_directory_path = cache_directory_path
                async with Site(configuration) as site:
                    site.ancestry.sources[resource.id] = resource
                    sut = _Populator(site, m_retriever)
                    await sut.populate()
        self.assertSetEqual({link}, resource.links)

    @patch('betty.plugin.wikipedia.Retriever', spec=Retriever, new_callable=AsyncMock)
    @sync
    async def test_populate_should_populate_existing_link(self, m_retriever) -> None:
        entry_language = 'en'
        entry_name = 'Amsterdam'
        entry_title = 'Amsterdam'
        entry_content = 'Capitol of the Netherlands'
        entry = Entry(entry_language, entry_name, entry_title, entry_content)
        m_retriever.get_entry.return_value = entry

        resource = IdentifiableSource('the_source', 'The Source')
        link = Link('https://en.wikipedia.org/wiki/Amsterdam')
        resource.links.add(link)
        with TemporaryDirectory() as output_directory_path:
            with TemporaryDirectory() as cache_directory_path:
                configuration = Configuration(
                    output_directory_path, 'https://example.com')
                configuration.cache_directory_path = cache_directory_path
                async with Site(configuration) as site:
                    site.ancestry.sources[resource.id] = resource
                    sut = _Populator(site, m_retriever)
                    await sut.populate()
        m_retriever.get_entry.assert_called_once_with(entry_language, entry_name)
        self.assertEqual(1, len(resource.links))
        self.assertEqual('Amsterdam', link.label)
        self.assertEqual('en', link.locale)
        self.assertEqual(MediaType('text/html'), link.media_type)
        self.assertIsNotNone(link.description)
        self.assertEqual('external', link.relationship)

    @patch('betty.plugin.wikipedia.Retriever', spec=Retriever, new_callable=AsyncMock)
    @sync
    async def test_populate_should_add_translation_links(self, m_retriever) -> None:
        entry_language = 'en'
        entry_name = 'Amsterdam'
        entry_title = 'Amsterdam'
        entry_content = 'Capitol of the Netherlands'
        entry = Entry(entry_language, entry_name, entry_title, entry_content)
        added_entry_language = 'nl'
        added_entry_name = 'Amsterdam'
        added_entry_title = 'Amsterdam'
        added_entry_content = 'Hoofdstad van Nederland'
        added_entry = Entry(added_entry_language, added_entry_name, added_entry_title, added_entry_content)
        m_retriever.get_entry.side_effect = [
            entry,
            added_entry
        ]

        m_retriever.get_translations.return_value = {
            entry_language: entry_name,
            added_entry_language: added_entry_name,
        }

        resource = IdentifiableSource('the_source', 'The Source')
        link_en = Link('https://en.wikipedia.org/wiki/Amsterdam')
        resource.links.add(link_en)
        with TemporaryDirectory() as output_directory_path:
            with TemporaryDirectory() as cache_directory_path:
                configuration = Configuration(
                    output_directory_path, 'https://example.com')
                configuration.cache_directory_path = cache_directory_path
                configuration.locales.clear()
                configuration.locales['en-US'] = LocaleConfiguration('en-US', 'en')
                configuration.locales['nl-NL'] = LocaleConfiguration('nl-NL', 'nl')
                async with Site(configuration) as site:
                    site.ancestry.sources[resource.id] = resource
                    sut = _Populator(site, m_retriever)
                    await sut.populate()

        m_retriever.get_entry.assert_has_calls([
            call(entry_language, entry_name),
            call(added_entry_language, added_entry_name),
        ])
        m_retriever.get_translations.assert_called_once_with(entry_language, entry_name)
        self.assertEqual(2, len(resource.links))
        link_nl = resource.links.difference({link_en}).pop()
        self.assertEqual('Amsterdam', link_nl.label)
        self.assertEqual('nl', link_nl.locale)
        self.assertEqual(MediaType('text/html'), link_nl.media_type)
        self.assertIsNotNone(link_nl.description)
        self.assertEqual('external', link_nl.relationship)


class WikipediaTest(TestCase):
    @aioresponses()
    @sync
    async def test_filter(self, m_aioresponses) -> None:
        entry_url = 'https://en.wikipedia.org/wiki/Amsterdam'
        links = [
            Link(entry_url),
            # Add a link to Wikipedia, but using a locale that's not used by the site, to test it's ignored.
            Link('https://nl.wikipedia.org/wiki/Amsterdam'),
            # Add a link that doesn't point to Wikipedia at all to test it's ignored.
            Link('https://example.com'),
        ]
        api_url = 'https://en.wikipedia.org/w/api.php?action=query&titles=Amsterdam&prop=extracts&exintro&format=json&formatversion=2'
        title = 'Amstelredam'
        extract = 'De hoofdstad van Nederland.'
        api_response_body = {
            'query': {
                'pages': [
                    {
                        'title': title,
                        'extract': extract,
                    },
                ],
            }
        }
        m_aioresponses.get(api_url, payload=api_response_body)

        with TemporaryDirectory() as output_directory_path:
            with TemporaryDirectory() as cache_directory_path:
                configuration = Configuration(
                    output_directory_path, 'https://ancestry.example.com')
                configuration.cache_directory_path = cache_directory_path
                configuration.plugins[Wikipedia] = None
                async with Site(configuration) as site:
                    actual = await site.jinja2_environment.from_string(
                        '{% for entry in (links | wikipedia) %}{{ entry.content }}{% endfor %}').render_async(links=links)
        self.assertEquals(extract, actual)

    @aioresponses()
    @sync
    async def test_post_parse(self, m_aioresponses) -> None:
        resource = IdentifiableSource('the_source', 'The Source')
        link = Link('https://en.wikipedia.org/wiki/Amsterdam')
        resource.links.add(link)
        entry_title = 'Amstelredam'
        entry_extract = 'Capitol of the Netherlands'
        entry_api_response_body = {
            'query': {
                'pages': [
                    {
                        'title': entry_title,
                        'extract': entry_extract,
                    },
                ],
            }
        }
        entry_api_url = 'https://en.wikipedia.org/w/api.php?action=query&titles=Amsterdam&prop=extracts&exintro&format=json&formatversion=2'
        m_aioresponses.get(entry_api_url, payload=entry_api_response_body)
        translations_api_response_body = {
            'query': {
                'pages': [
                    {
                        'langlinks': [],
                    },
                ],
            },
        }
        translations_api_url = 'https://en.wikipedia.org/w/api.php?action=query&titles=Amsterdam&prop=langlinks&lllimit=500&format=json&formatversion=2'
        m_aioresponses.get(translations_api_url, payload=translations_api_response_body)

        with TemporaryDirectory() as output_directory_path:
            with TemporaryDirectory() as cache_directory_path:
                configuration = Configuration(
                    output_directory_path, 'https://example.com')
                configuration.cache_directory_path = cache_directory_path
                configuration.plugins[Wikipedia] = None
                async with Site(configuration) as site:
                    site.ancestry.sources[resource.id] = resource
                    await parse(site)

        self.assertEqual(1, len(resource.links))
        self.assertEqual(entry_title, link.label)
        self.assertEqual('en', link.locale)
        self.assertEqual(MediaType('text/html'), link.media_type)
        self.assertIsNotNone(link.description)
        self.assertEqual('external', link.relationship)



import asyncio
import inspect
from contextlib import suppress
from functools import wraps


def _wrap_sync(f):
    @wraps(f)
    def _synced(*args, **kwargs):
        return sync(f(*args, **kwargs))
    return _synced


def sync(f):
    if inspect.iscoroutine(f):
        return asyncio.get_event_loop().run_until_complete(f)

    if inspect.iscoroutinefunction(f):
        return _wrap_sync(f)

    if callable(f):
        with suppress(AttributeError):
            if inspect.iscoroutinefunction(getattr(f, '__call__')):
                return _wrap_sync(f)
        return f

    raise ValueError('Can only synchronize coroutine callables (`async def`) or coroutines (values returned by `async def`), or pass through synchronous callables, but "%s" was given.' % f)



from betty.ancestry import PlaceName, Place, Enclosure
from betty.asyncio import sync
from betty.tests import TemplateTestCase


class Test(TemplateTestCase):
    template_file = 'meta/place.html.j2'

    @sync
    async def test_without_enclosing_places(self):
        place = Place('P0', [PlaceName('The Place')])
        expected = '<div class="meta"></div>'
        async with self._render(data={
            'place': place,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_enclosing_place_without_place_context(self):
        place = Place('P0', [PlaceName('The Place')])
        enclosing_place = Place('P1', [PlaceName('The Enclosing Place')])
        Enclosure(place, enclosing_place)
        all_enclosing_place = Place('P2', [PlaceName('The All-enclosing Place')])
        Enclosure(enclosing_place, all_enclosing_place)
        expected = '<div class="meta">in <address><a href="/place/P1/index.html"><span>The Enclosing Place</span></a></address>, <address><a href="/place/P2/index.html"><span>The All-enclosing Place</span></a></address></div>'
        async with self._render(data={
            'place': place,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_enclosing_place_with_matching_place_context(self):
        place = Place('P0', [PlaceName('The Place')])
        enclosing_place = Place('P1', [PlaceName('The Enclosing Place')])
        Enclosure(place, enclosing_place)
        all_enclosing_place = Place('P2', [PlaceName('The All-enclosing Place')])
        Enclosure(enclosing_place, all_enclosing_place)
        expected = '<div class="meta">in <address><a href="/place/P1/index.html"><span>The Enclosing Place</span></a></address></div>'
        async with self._render(data={
            'place': place,
            'place_context': all_enclosing_place,
        }) as (actual, _):
            self.assertEqual(expected, actual)

    @sync
    async def test_with_enclosing_place_with_non_matching_place_context(self):
        place = Place('P0', [PlaceName('The Place')])
        enclosing_place = Place('P1', [PlaceName('The Enclosing Place')])
        Enclosure(place, enclosing_place)
        all_enclosing_place = Place('P2', [PlaceName('The All-enclosing Place')])
        Enclosure(enclosing_place, all_enclosing_place)
        unrelated_place = Place('P999', [PlaceName('Far Far Away')])
        expected = '<div class="meta">in <address><a href="/place/P1/index.html"><span>The Enclosing Place</span></a></address>, <address><a href="/place/P2/index.html"><span>The All-enclosing Place</span></a></address></div>'
        async with self._render(data={
            'place': place,
            'place_context': unrelated_place,
        }) as (actual, _):
            self.assertEqual(expected, actual)
